# Copyright 2013-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import threading

from portage import os
from portage.checksum import (
    _apply_hash_filter,
    _filter_unaccelarated_hashes,
    _hash_filter,
)
from portage.dep import use_reduce
from portage.exception import PortageException, PortageKeyError
from portage.package.ebuild.fetch import DistfileName
from portage.util._async.AsyncTaskFuture import AsyncTaskFuture
from portage.util._async.TaskScheduler import TaskScheduler
from portage.util.futures import asyncio
from portage.util.futures.iter_completed import iter_gather
from .FetchTask import FetchTask
from _emerge.CompositeTask import CompositeTask


class FetchIterator:
    def __init__(self, config):
        self._config = config
        self._terminated = threading.Event()

    def terminate(self):
        """
        Schedules early termination of the __iter__ method, which is
        useful because under some conditions it's possible for __iter__
        to loop for a long time without yielding to the caller. For
        example, it's useful when there are many ebuilds with stale
        cache and RESTRICT=mirror.

        This method is thread-safe (and safe for signal handlers).
        """
        self._terminated.set()

    def _iter_every_cp(self):
        # List categories individually, in order to start yielding quicker,
        # and in order to reduce latency in case of a signal interrupt.
        cp_all = self._config.portdb.cp_all
        for category in sorted(self._config.portdb.categories):
            yield from cp_all(categories=(category,))

    def __iter__(self):
        portdb = self._config.portdb
        get_repo_for_location = portdb.repositories.get_repo_for_location

        hash_filter = _hash_filter(portdb.settings.get("PORTAGE_CHECKSUM_FILTER", ""))
        if hash_filter.transparent:
            hash_filter = None

        for cp in self._iter_every_cp():
            if self._terminated.is_set():
                return

            for tree in portdb.porttrees:
                # Reset state so the Manifest is pulled once
                # for this cp / tree combination.
                repo_config = get_repo_for_location(tree)
                digests_future = portdb._event_loop.create_future()

                for cpv in portdb.cp_list(cp, mytree=tree):
                    if self._terminated.is_set():
                        return

                    yield _EbuildFetchTasks(
                        fetch_tasks_future=_async_fetch_tasks(
                            self._config,
                            hash_filter,
                            repo_config,
                            digests_future,
                            cpv,
                            portdb._event_loop,
                        )
                    )


class _EbuildFetchTasks(CompositeTask):
    """
    Executes FetchTask instances (which are asynchronously constructed)
    for each of the files referenced by an ebuild.
    """

    __slots__ = ("fetch_tasks_future",)

    def _start(self):
        self._start_task(
            AsyncTaskFuture(future=self.fetch_tasks_future), self._start_fetch_tasks
        )

    def _start_fetch_tasks(self, task):
        if self._default_exit(task) != os.EX_OK:
            self._async_wait()
            return

        self._start_task(
            TaskScheduler(
                iter(self.fetch_tasks_future.result()),
                max_jobs=1,
                event_loop=self.scheduler,
            ),
            self._default_final_exit,
        )


def _async_fetch_tasks(config, hash_filter, repo_config, digests_future, cpv, loop):
    """
    Asynchronously construct FetchTask instances for each of the files
    referenced by an ebuild.

    @param config: emirrordist config
    @type config: portage._emirrordist.Config.Config
    @param hash_filter: PORTAGE_CHECKSUM_FILTER settings
    @type hash_filter: portage.checksum._hash_filter
    @param repo_config: repository configuration
    @type repo_config: RepoConfig
    @param digests_future: future that contains cached distfiles digests
            for the current cp if available
    @type digests_future: asyncio.Future
    @param cpv: current ebuild cpv
    @type cpv: portage.versions._pkg_str
    @param loop: event loop
    @type loop: EventLoop
    @return: A future that results in a list containing FetchTask
            instances for each of the files referenced by an ebuild.
    @rtype: asyncio.Future (or compatible)
    """
    result = loop.create_future()
    fetch_tasks = []

    def aux_get_done(gather_result):
        # All exceptions must be consumed from gather_result before this
        # function returns, in order to avoid triggering the event loop's
        # exception handler.
        if not gather_result.cancelled():
            list(
                future.exception()
                for future in gather_result.result()
                if not future.cancelled()
            )
        else:
            result.cancel()

        if result.cancelled():
            return

        aux_get_result, fetch_map_result = gather_result.result()
        if aux_get_result.cancelled() or fetch_map_result.cancelled():
            # Cancel result after consuming any exceptions which
            # are now irrelevant due to cancellation.
            aux_get_result.cancelled() or aux_get_result.exception()
            fetch_map_result.cancelled() or fetch_map_result.exception()
            result.cancel()
            return

        try:
            (restrict,) = aux_get_result.result()
        except (PortageKeyError, PortageException) as e:
            config.log_failure(f"{cpv}\t\taux_get exception {e}")
            result.set_result(fetch_tasks)
            return

        # Here we use matchnone=True to ignore conditional parts
        # of RESTRICT since they don't apply unconditionally.
        # Assume such conditionals only apply on the client side.
        try:
            restrict = frozenset(use_reduce(restrict, flat=True, matchnone=True))
        except PortageException as e:
            config.log_failure(f"{cpv}\t\tuse_reduce exception {e}")
            result.set_result(fetch_tasks)
            return

        try:
            uri_map = fetch_map_result.result()
        except PortageException as e:
            config.log_failure(f"{cpv}\t\tgetFetchMap exception {e}")
            result.set_result(fetch_tasks)
            return

        if not uri_map:
            result.set_result(fetch_tasks)
            return

        new_uri_map = {}
        restrict_fetch = "fetch" in restrict
        restrict_mirror = restrict_fetch or "mirror" in restrict
        for filename, uri_tuple in uri_map.items():
            new_uris = []
            for uri in uri_tuple:
                override_mirror = uri.startswith("mirror+")
                override_fetch = override_mirror or uri.startswith("fetch+")
                if override_fetch:
                    uri = uri.partition("+")[2]

                # skip fetch-restricted files unless overridden via fetch+
                # or mirror+
                if restrict_fetch and not override_fetch:
                    continue
                # skip mirror-restricted files unless override via mirror+
                # or in config_mirror_exemptions
                if restrict_mirror and not override_mirror:
                    if config.restrict_mirror_exemptions is None or not uri.startswith(
                        "mirror://"
                    ):
                        continue
                    mirror_name = uri.split("/", 3)[2]
                    if mirror_name not in config.restrict_mirror_exemptions:
                        continue
                # if neither fetch or mirror restriction applies to the URI
                # or it is exempted from them, readd it (with fetch+/mirror+
                # prefix stripped)
                new_uris.append(uri)

            # if we've gotten any new URIs, then we readd the file
            if new_uris:
                new_uri_map[filename] = new_uris

        if not new_uri_map:
            result.set_result(fetch_tasks)
            return

        # Parse Manifest for this cp if we haven't yet.
        try:
            if digests_future.done():
                # If there's an exception then raise it.
                digests = digests_future.result()
            else:
                digests = repo_config.load_manifest(
                    os.path.join(repo_config.location, cpv.cp)
                ).getTypeDigests("DIST")
        except (OSError, PortageException) as e:
            digests_future.done() or digests_future.set_exception(e)
            for filename in new_uri_map:
                config.log_failure(f"{cpv}\t{filename}\tManifest exception {e}")
                config.file_failures[filename] = cpv
            result.set_result(fetch_tasks)
            return
        else:
            digests_future.done() or digests_future.set_result(digests)

        if not digests:
            for filename in new_uri_map:
                config.log_failure(f"{cpv}\t{filename}\tdigest entry missing")
                config.file_failures[filename] = cpv
            result.set_result(fetch_tasks)
            return

        for filename, uri_tuple in new_uri_map.items():
            file_digests = digests.get(filename)
            if file_digests is None:
                config.log_failure(f"{cpv}\t{filename}\tdigest entry missing")
                config.file_failures[filename] = cpv
                continue
            if filename in config.file_owners:
                continue
            config.file_owners[filename] = cpv

            file_digests = _filter_unaccelarated_hashes(file_digests)
            if hash_filter is not None:
                file_digests = _apply_hash_filter(file_digests, hash_filter)

            fetch_tasks.append(
                FetchTask(
                    cpv=cpv,
                    background=True,
                    digests=file_digests,
                    distfile=DistfileName(filename, digests=file_digests),
                    restrict=restrict,
                    uri_tuple=uri_tuple,
                    config=config,
                )
            )

        result.set_result(fetch_tasks)

    def future_generator():
        yield asyncio.ensure_future(
            config.portdb.async_aux_get(
                cpv, ("RESTRICT",), myrepo=repo_config.name, loop=loop
            ),
            loop,
        )
        yield config.portdb.async_fetch_map(cpv, mytree=repo_config.location, loop=loop)

    # Use iter_gather(max_jobs=1) to limit the number of processes per
    # _EbuildFetchTask instance, and also to avoid spawning two bash
    # processes for the same cpv simultaneously (the second one can
    # use metadata cached by the first one).
    gather_result = iter_gather(
        future_generator(),
        max_jobs=1,
        loop=loop,
    )
    gather_result.add_done_callback(aux_get_done)
    result.add_done_callback(
        lambda result: (
            gather_result.cancel()
            if result.cancelled() and not gather_result.done()
            else None
        )
    )

    return result



# SPDX-License-Identifier: Apache-2.0
# Copyright 2017 The Meson development team
# Copyright © 2023-2025 Intel Corporation

from __future__ import annotations


import abc
import argparse
import os
import sys
import shlex
import shutil
import subprocess
import tarfile
import tempfile
import hashlib
import typing as T

from dataclasses import dataclass
from glob import glob
from pathlib import Path
from mesonbuild.environment import Environment, detect_ninja
from mesonbuild.mesonlib import (GIT, MesonException, RealPathAction, get_meson_command, quiet_git,
                                 windows_proof_rmtree, setup_vsenv)
from .options import OptionKey
from mesonbuild.msetup import add_arguments as msetup_argparse
from mesonbuild.wrap import wrap
from mesonbuild import mlog, build, coredata
from .scripts.meson_exe import run_exe

if T.TYPE_CHECKING:
    from ._typing import ImmutableListProtocol
    from .mesonlib import ExecutableSerialisation

archive_choices = ['bztar', 'gztar', 'xztar', 'zip']

archive_extension = {'bztar': '.tar.bz2',
                     'gztar': '.tar.gz',
                     'xztar': '.tar.xz',
                     'zip': '.zip'}

if sys.version_info >= (3, 14):
    tarfile.TarFile.extraction_filter = staticmethod(tarfile.fully_trusted_filter)

# Note: when adding arguments, please also add them to the completion
# scripts in $MESONSRC/data/shell-completions/
def add_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument('-C', dest='wd', action=RealPathAction,
                        help='directory to cd into before running')
    parser.add_argument('--allow-dirty', action='store_true',
                        help='Allow even when repository contains uncommitted changes.')
    parser.add_argument('--formats', default='xztar',
                        help='Comma separated list of archive types to create. Supports xztar (default), bztar, gztar, and zip.')
    parser.add_argument('--include-subprojects', action='store_true',
                        help='Include source code of subprojects that have been used for the build.')
    parser.add_argument('--no-tests', action='store_true',
                        help='Do not build and test generated packages.')


def create_hash(fname: str) -> None:
    hashname = fname + '.sha256sum'
    m = hashlib.sha256()
    m.update(open(fname, 'rb').read())
    with open(hashname, 'w', encoding='utf-8') as f:
        # A space and an asterisk because that is the format defined by GNU coreutils
        # and accepted by busybox and the Perl shasum tool.
        f.write('{} *{}\n'.format(m.hexdigest(), os.path.basename(fname)))


msg_uncommitted_changes = 'Repository has uncommitted changes that will not be included in the dist tarball'

def handle_dirty_opt(msg: str, allow_dirty: bool) -> None:
    if allow_dirty:
        mlog.warning(msg)
    else:
        mlog.error(msg + '\n' + 'Use --allow-dirty to ignore the warning and proceed anyway')
        sys.exit(1)

def is_git(src_root: str) -> bool:
    '''
    Checks if meson.build file at the root source directory is tracked by git.
    It could be a subproject part of the parent project git repository.
    '''
    if quiet_git(['ls-files', '--error-unmatch', 'meson.build'], src_root)[0]:
        return True

    if os.path.exists(os.path.join(src_root, '.git')):
        msg = 'Source tree looks like it may be a git repo, '
        if not GIT:
            msg += 'but git is not installed!'
            if 'GITLAB_CI' in os.environ:
                msg += ' This is a gitlab bug.'
        else:
            msg += 'but git returned a failure. '
            p, oe = quiet_git(['status'], src_root)
            if 'dubious ownership' in oe:
                # For a few years now, git has absolved itself of the responsibility to implement
                # robust, safe software. Instead of detecting the signs of a problematic scenario,
                # they have chosen to consider many legitimate and reasonable use cases as "dangerous",
                # and implemented the number one threat to security worldwide: alert fatigue. Having
                # done so, they then washed their hands of the matter and permanently tabled the
                # notion of adding fine-grained detection. This is not just useless, it is *worse*
                # than useless.
                #
                # In our case, the error is triply meaningless since we are already executing build
                # system commands from the same directory. Either way, reject the notion that git is
                # well designed or that its error messaging is a valid approach to the problem space.
                msg += 'This is a bug in git itself, please set `git config --global safe.directory "*"`'
            else:
                msg += 'meson.build may not have been committed to git?'
        mlog.warning(msg)
    return False


def is_hg(src_root: str) -> bool:
    return os.path.isdir(os.path.join(src_root, '.hg'))


@dataclass
class Dist(metaclass=abc.ABCMeta):
    dist_name: str
    src_root: str
    bld_root: str
    dist_scripts: T.List[ExecutableSerialisation]
    subprojects: T.Dict[str, str]
    options: argparse.Namespace

    def __post_init__(self) -> None:
        self.dist_sub = os.path.join(self.bld_root, 'meson-dist')
        self.distdir = os.path.join(self.dist_sub, self.dist_name)

    @abc.abstractmethod
    def create_dist(self, archives: T.List[str]) -> T.List[str]:
        pass

    def run_dist_scripts(self) -> None:
        assert os.path.isabs(self.distdir)
        mesonrewrite = Environment.get_build_command() + ['rewrite']
        env = {'MESON_DIST_ROOT': self.distdir,
               'MESON_SOURCE_ROOT': self.src_root,
               'MESON_BUILD_ROOT': self.bld_root,
               'MESONREWRITE': ' '.join(shlex.quote(x) for x in mesonrewrite),
               }
        for d in self.dist_scripts:
            if d.subproject and d.subproject not in self.subprojects:
                continue
            subdir = self.subprojects.get(d.subproject, '')
            env['MESON_PROJECT_DIST_ROOT'] = os.path.join(self.distdir, subdir)
            env['MESON_PROJECT_SOURCE_ROOT'] = os.path.join(self.src_root, subdir)
            env['MESON_PROJECT_BUILD_ROOT'] = os.path.join(self.bld_root, subdir)
            name = ' '.join(d.cmd_args)
            print(f'Running custom dist script {name!r}')
            try:
                rc = run_exe(d, env)
                if rc != 0:
                    sys.exit('Dist script errored out')
            except OSError:
                print(f'Failed to run dist script {name!r}')
                sys.exit(1)


class GitDist(Dist):
    def git_root(self, dir_: str) -> Path:
        # Cannot use --show-toplevel here because git in our CI prints cygwin paths
        # that python cannot resolve. Workaround this by taking parent of src_root.
        prefix = quiet_git(['rev-parse', '--show-prefix'], dir_, check=True)[1].strip()
        if not prefix:
            return Path(dir_)
        prefix_level = len(Path(prefix).parents)
        return Path(dir_).parents[prefix_level - 1]

    def have_dirty_index(self) -> bool:
        '''Check whether there are uncommitted changes in git'''
        # Optimistically call update-index, and disregard its return value. It could be read-only,
        # and only the output of diff-index matters.
        subprocess.call(['git', '-C', self.src_root, 'update-index', '-q', '--refresh'])
        ret = subprocess.call(['git', '-C', self.src_root, 'diff-index', '--quiet', 'HEAD'])
        return ret == 1

    def copy_git(self, src: T.Union[str, os.PathLike], distdir: str, revision: str = 'HEAD',
                 prefix: T.Optional[str] = None, subdir: T.Optional[str] = None) -> None:
        cmd = ['git', 'archive', '--format', 'tar', revision]
        if prefix is not None:
            cmd.insert(2, f'--prefix={prefix}/')
        if subdir is not None:
            cmd.extend(['--', subdir])
        with tempfile.TemporaryFile() as f:
            subprocess.check_call(cmd, cwd=src, stdout=f)
            f.seek(0)
            t = tarfile.open(fileobj=f) # [ignore encoding]
            t.extractall(path=distdir)

    def process_git_project(self, src_root: str, distdir: str) -> None:
        if self.have_dirty_index():
            handle_dirty_opt(msg_uncommitted_changes, self.options.allow_dirty)
        if os.path.exists(distdir):
            windows_proof_rmtree(distdir)
        repo_root = self.git_root(src_root)
        if repo_root.samefile(src_root):
            os.makedirs(distdir)
            self.copy_git(src_root, distdir)
        else:
            subdir = Path(src_root).relative_to(repo_root)
            tmp_distdir = distdir + '-tmp'
            if os.path.exists(tmp_distdir):
                windows_proof_rmtree(tmp_distdir)
            os.makedirs(tmp_distdir)
            self.copy_git(repo_root, tmp_distdir, subdir=str(subdir))
            Path(tmp_distdir, subdir).rename(distdir)
            windows_proof_rmtree(tmp_distdir)
        self.process_submodules(src_root, distdir)

    def process_submodules(self, src: str, distdir: str) -> None:
        module_file = os.path.join(src, '.gitmodules')
        if not os.path.exists(module_file):
            return
        cmd = ['git', 'submodule', 'status', '--cached', '--recursive']
        modlist = subprocess.check_output(cmd, cwd=src, universal_newlines=True).splitlines()
        for submodule in modlist:
            status = submodule[:1]
            sha1, rest = submodule[1:].split(' ', 1)
            subpath = rest.rsplit(' ', 1)[0]

            if status == '-':
                mlog.warning(f'Submodule {subpath!r} is not checked out and cannot be added to the dist')
                continue
            elif status in {'+', 'U'}:
                handle_dirty_opt(f'Submodule {subpath!r} has uncommitted changes that will not be included in the dist tarball', self.options.allow_dirty)

            self.copy_git(os.path.join(src, subpath), distdir, revision=sha1, prefix=subpath)

    def create_dist(self, archives: T.List[str]) -> T.List[str]:
        self.process_git_project(self.src_root, self.distdir)
        for path in self.subprojects.values():
            sub_src_root = os.path.join(self.src_root, path)
            sub_distdir = os.path.join(self.distdir, path)
            if os.path.exists(sub_distdir):
                continue
            if is_git(sub_src_root):
                self.process_git_project(sub_src_root, sub_distdir)
            else:
                shutil.copytree(sub_src_root, sub_distdir)
        self.run_dist_scripts()
        output_names = []
        for a in archives:
            compressed_name = self.distdir + archive_extension[a]
            shutil.make_archive(self.distdir, a, root_dir=self.dist_sub, base_dir=self.dist_name)
            output_names.append(compressed_name)
        windows_proof_rmtree(self.distdir)
        return output_names


class HgDist(Dist):
    def have_dirty_index(self) -> bool:
        '''Check whether there are uncommitted changes in hg'''
        env = os.environ.copy()
        env['LC_ALL'] = 'C'
        # cpython's gettext has a bug and uses LANGUAGE to override LC_ALL,
        # contrary to the gettext spec
        env.pop('LANGUAGE', None)
        out = subprocess.check_output(['hg', '-R', self.src_root, 'summary'], env=env)
        return b'commit: (clean)' not in out

    def create_dist(self, archives: T.List[str]) -> T.List[str]:
        if self.have_dirty_index():
            handle_dirty_opt(msg_uncommitted_changes, self.options.allow_dirty)
        if self.dist_scripts:
            mlog.warning('dist scripts are not supported in Mercurial projects')

        os.makedirs(self.dist_sub, exist_ok=True)
        tarname = os.path.join(self.dist_sub, self.dist_name + '.tar')
        xzname = tarname + '.xz'
        bz2name = tarname + '.bz2'
        gzname = tarname + '.gz'
        zipname = os.path.join(self.dist_sub, self.dist_name + '.zip')
        # Note that -X interprets relative paths using the current working
        # directory, not the repository root, so this must be an absolute path:
        # https://bz.mercurial-scm.org/show_bug.cgi?id=6267
        #
        # .hg[a-z]* is used instead of .hg* to keep .hg_archival.txt, which may
        # be useful to link the tarball to the Mercurial revision for either
        # manual inspection or in case any code interprets it for a --version or
        # similar.
        subprocess.check_call(['hg', 'archive', '-R', self.src_root, '-S', '-t', 'tar',
                               '-X', self.src_root + '/.hg[a-z]*', tarname])
        output_names = []
        if 'xztar' in archives:
            import lzma
            with lzma.open(xzname, 'wb') as xf, open(tarname, 'rb') as tf:
                shutil.copyfileobj(tf, xf)
            output_names.append(xzname)
        if 'bztar' in archives:
            import bz2
            with bz2.open(bz2name, 'wb') as bf, open(tarname, 'rb') as tf:
                shutil.copyfileobj(tf, bf)
            output_names.append(bz2name)
        if 'gztar' in archives:
            import gzip
            with gzip.open(gzname, 'wb') as zf, open(tarname, 'rb') as tf:
                shutil.copyfileobj(tf, zf)
            output_names.append(gzname)
        os.unlink(tarname)
        if 'zip' in archives:
            subprocess.check_call(['hg', 'archive', '-R', self.src_root, '-S', '-t', 'zip', zipname])
            output_names.append(zipname)
        return output_names


def run_dist_steps(meson_command: T.List[str], unpacked_src_dir: str, builddir: str, installdir: str, ninja_args: T.List[str]) -> int:
    if subprocess.call(meson_command + ['--backend=ninja', unpacked_src_dir, builddir]) != 0:
        print('Running Meson on distribution package failed')
        return 1
    if subprocess.call(ninja_args, cwd=builddir) != 0:
        print('Compiling the distribution package failed')
        return 1
    if subprocess.call(ninja_args + ['test'], cwd=builddir) != 0:
        print('Running unit tests on the distribution package failed')
        return 1
    myenv = os.environ.copy()
    myenv['DESTDIR'] = installdir
    if subprocess.call(ninja_args + ['install'], cwd=builddir, env=myenv) != 0:
        print('Installing the distribution package failed')
        return 1
    return 0

def check_dist(packagename: str, _meson_command: ImmutableListProtocol[str], extra_meson_args: T.List[str], bld_root: str, privdir: str) -> int:
    print(f'Testing distribution package {packagename}')
    unpackdir = os.path.join(privdir, 'dist-unpack')
    builddir = os.path.join(privdir, 'dist-build')
    installdir = os.path.join(privdir, 'dist-install')
    for p in (unpackdir, builddir, installdir):
        if os.path.exists(p):
            windows_proof_rmtree(p)
        os.mkdir(p)
    ninja_args = detect_ninja()
    shutil.unpack_archive(packagename, unpackdir)
    unpacked_files = glob(os.path.join(unpackdir, '*'))
    assert len(unpacked_files) == 1
    unpacked_src_dir = unpacked_files[0]
    meson_command = _meson_command.copy()
    meson_command += ['setup']
    meson_command += create_cmdline_args(bld_root)
    meson_command += extra_meson_args

    ret = run_dist_steps(meson_command, unpacked_src_dir, builddir, installdir, ninja_args)
    if ret > 0:
        print(f'Dist check build directory was {builddir}')
    else:
        windows_proof_rmtree(unpackdir)
        windows_proof_rmtree(builddir)
        windows_proof_rmtree(installdir)
        print(f'Distribution package {packagename} tested')
    return ret

def create_cmdline_args(bld_root: str) -> T.List[str]:
    parser = argparse.ArgumentParser()
    msetup_argparse(parser)
    args = T.cast('coredata.SharedCMDOptions', parser.parse_args([]))
    coredata.parse_cmd_line_options(args)
    coredata.read_cmd_line_file(bld_root, args)
    args.cmd_line_options.pop(OptionKey('backend'), '')
    return shlex.split(coredata.format_cmd_line_options(args))

def determine_archives_to_generate(options: argparse.Namespace) -> T.List[str]:
    result = []
    for i in options.formats.split(','):
        if i not in archive_choices:
            sys.exit(f'Value "{i}" not one of permitted values {archive_choices}.')
        result.append(i)
    if len(i) == 0:
        sys.exit('No archive types specified.')
    return result

def run(options: argparse.Namespace) -> int:
    buildfile = Path(options.wd) / 'meson-private' / 'build.dat'
    if not buildfile.is_file():
        raise MesonException(f'Directory {options.wd!r} does not seem to be a Meson build directory.')
    b = build.load(options.wd)
    need_vsenv = T.cast('bool', b.environment.coredata.optstore.get_value_for(OptionKey('vsenv')))
    setup_vsenv(need_vsenv)
    src_root = b.environment.source_dir
    bld_root = b.environment.build_dir
    priv_dir = os.path.join(bld_root, 'meson-private')

    dist_name = b.project_name + '-' + b.project_version

    archives = determine_archives_to_generate(options)

    subprojects = {}
    extra_meson_args = []
    if options.include_subprojects:
        subproject_dir = os.path.join(src_root, b.subproject_dir)
        for sub in b.subprojects:
            directory = wrap.get_directory(subproject_dir, sub)
            subprojects[sub] = os.path.join(b.subproject_dir, directory)
        extra_meson_args.append('-Dwrap_mode=nodownload')

    cls: T.Type[Dist]
    if is_git(src_root):
        cls = GitDist
    elif is_hg(src_root):
        if subprojects:
            print('--include-subprojects option currently not supported with Mercurial')
            return 1
        cls = HgDist
    else:
        print('Dist currently only works with Git or Mercurial repos')
        return 1

    project = cls(dist_name, src_root, bld_root, b.dist_scripts, subprojects, options)
    names = project.create_dist(archives)

    if names is None:
        return 1
    rc = 0
    if not options.no_tests:
        # Check only one.
        rc = check_dist(names[0], get_meson_command(), extra_meson_args, bld_root, priv_dir)
    if rc == 0:
        for name in names:
            create_hash(name)
            print('Created', name)
    return rc



# Copyright 2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class ProfileUseStableTestCase(TestCase):
    def testProfileUseStable(self):
        profile = {
            "eapi": ("9-pre1",),
            "package.use": ("app-misc/A -a d",),
            "package.use.stable": ("app-misc/A a -b -c f",),
            "use.stable": ("a", "b", "c", "e"),
        }

        user_config = {
            "package.accept_keywords": ("=app-misc/A-2 ~x86",),
        }

        ebuilds = {
            "app-misc/A-1": {"EAPI": "8", "KEYWORDS": "x86", "IUSE": "a b c d e f"},
            "app-misc/A-2": {"EAPI": "8", "KEYWORDS": "~x86", "IUSE": "a b c d e f"},
            "app-misc/B-1": {
                "EAPI": "8",
                # package.use.stable > package.use > use.stable
                "RDEPEND": "=app-misc/A-1[a,-b,-c,d,e,f]",
            },
            "app-misc/C-1": {
                "EAPI": "8",
                # package.use.stable and use.stable do not apply due to unstable keyword
                "RDEPEND": "=app-misc/A-2[-a,-b,-c,d,-e,-f]",
            },
        }

        test_cases = (
            # Test stable package
            ResolverPlaygroundTestCase(
                ["app-misc/B"],
                success=True,
                mergelist=[
                    "app-misc/A-1",
                    "app-misc/B-1",
                ],
            ),
            # Test unstable package
            ResolverPlaygroundTestCase(
                ["app-misc/C"],
                success=True,
                mergelist=[
                    "app-misc/A-2",
                    "app-misc/C-1",
                ],
            ),
        )

        playground = ResolverPlayground(
            debug=False,
            ebuilds=ebuilds,
            profile=profile,
            user_config=user_config,
        )

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            # Disable debug so that cleanup works.
            playground.debug = False
            playground.cleanup()



# SPDX-License-Identifier: Apache-2.0
# Copyright 2012-2022 The Meson development team
# Copyright © 2023-2025 Intel Corporation

from __future__ import annotations

import functools
import os.path
import textwrap
import re
import typing as T

from .. import options
from ..mesonlib import EnvironmentException, MesonException, Popen_safe_logged, version_compare
from ..options import OptionKey
from .compilers import Compiler, CompileCheckMode, clike_debug_args, is_library

if T.TYPE_CHECKING:
    from ..options import MutableKeyedOptionDictType
    from ..envconfig import MachineInfo
    from ..environment import Environment  # noqa: F401
    from ..linkers.linkers import DynamicLinker
    from ..mesonlib import MachineChoice
    from ..dependencies import Dependency
    from ..build import BuildTarget


rust_optimization_args: T.Dict[str, T.List[str]] = {
    'plain': [],
    '0': [],
    'g': ['-C', 'opt-level=0'],
    '1': ['-C', 'opt-level=1'],
    '2': ['-C', 'opt-level=2'],
    '3': ['-C', 'opt-level=3'],
    's': ['-C', 'opt-level=s'],
}

def get_rustup_run_and_args(exelist: T.List[str]) -> T.Optional[T.Tuple[T.List[str], T.List[str]]]:
    """Given the command for a rustc executable, check if it is invoked via
       "rustup run" and if so separate the "rustup [OPTIONS] run TOOLCHAIN"
       part from the arguments to rustc.  If the returned value is not None,
       other tools (for example clippy-driver or rustdoc) can be run by placing
       the name of the tool between the two elements of the tuple."""
    e = iter(exelist)
    try:
        if os.path.basename(next(e)) != 'rustup':
            return None
        # minimum three strings: "rustup run TOOLCHAIN"
        n = 3
        opt = next(e)

        # options come first
        while opt.startswith('-'):
            n += 1
            opt = next(e)

        # then "run TOOLCHAIN"
        if opt != 'run':
            return None

        next(e)
        next(e)
        return exelist[:n], list(e)
    except StopIteration:
        return None

class RustCompiler(Compiler):

    # rustc doesn't invoke the compiler itself, it doesn't need a LINKER_PREFIX
    language = 'rust'
    id = 'rustc'

    _WARNING_LEVELS: T.Dict[str, T.List[str]] = {
        '0': ['--cap-lints', 'allow'],
        '1': [],
        '2': [],
        '3': ['-W', 'warnings'],
        'everything': ['-W', 'warnings'],
    }

    # Those are static libraries, but we use dylib= here as workaround to avoid
    # rust --tests to use /WHOLEARCHIVE.
    # https://github.com/rust-lang/rust/issues/116910
    MSVCRT_ARGS: T.Mapping[str, T.List[str]] = {
        'none': [],
        'md': [], # this is the default, no need to inject anything
        'mdd': ['-l', 'dylib=msvcrtd'],
        'mt': ['-l', 'dylib=libcmt'],
        'mtd': ['-l', 'dylib=libcmtd'],
    }

    def __init__(self, exelist: T.List[str], version: str, for_machine: MachineChoice,
                 is_cross: bool, info: 'MachineInfo',
                 full_version: T.Optional[str] = None,
                 linker: T.Optional['DynamicLinker'] = None):
        super().__init__([], exelist, version, for_machine, info,
                         is_cross=is_cross, full_version=full_version,
                         linker=linker)
        self.rustup_run_and_args: T.Optional[T.Tuple[T.List[str], T.List[str]]] = get_rustup_run_and_args(exelist)
        self.base_options.update({OptionKey(o) for o in ['b_colorout', 'b_ndebug']})
        if 'link' in self.linker.id:
            self.base_options.add(OptionKey('b_vscrt'))
        self.native_static_libs: T.List[str] = []
        self.is_beta = '-beta' in full_version
        self.is_nightly = '-nightly' in full_version

    def needs_static_linker(self) -> bool:
        return False

    def sanity_check(self, work_dir: str, environment: Environment) -> None:
        source_name = os.path.join(work_dir, 'sanity.rs')
        output_name = os.path.join(work_dir, 'rusttest.exe')
        cmdlist = self.exelist.copy()

        with open(source_name, 'w', encoding='utf-8') as ofile:
            # If machine kernel is not `none`, try to compile a dummy program.
            # If 'none', this is likely a `no-std`(i.e. bare metal) project.
            if self.info.kernel != 'none':
                ofile.write(textwrap.dedent(
                    '''fn main() {
                    }
                    '''))
            else:
                # If rustc linker is gcc, add `-nostartfiles`
                if 'ld.' in self.linker.id:
                    cmdlist.extend(['-C', 'link-arg=-nostartfiles'])
                ofile.write(textwrap.dedent(
                    '''#![no_std]
                    #![no_main]
                    #[no_mangle]
                    pub fn _start() {
                    }
                    #[panic_handler]
                    fn panic(_info: &core::panic::PanicInfo) -> ! {
                        loop {}
                    }
                    '''))

        cmdlist.extend(['-o', output_name, source_name])
        pc, stdo, stde = Popen_safe_logged(cmdlist, cwd=work_dir)
        if pc.returncode != 0:
            raise EnvironmentException(f'Rust compiler {self.name_string()} cannot compile programs.')
        self._native_static_libs(work_dir, source_name)
        self.run_sanity_check(environment, [output_name], work_dir)

    def _native_static_libs(self, work_dir: str, source_name: str) -> None:
        # Get libraries needed to link with a Rust staticlib
        cmdlist = self.exelist + ['--crate-type', 'staticlib', '--print', 'native-static-libs', source_name]
        p, stdo, stde = Popen_safe_logged(cmdlist, cwd=work_dir)
        if p.returncode != 0:
            raise EnvironmentException('Rust compiler cannot compile staticlib.')
        match = re.search('native-static-libs: (.*)$', stde, re.MULTILINE)
        if not match:
            if self.info.kernel == 'none':
                # no match and kernel == none (i.e. baremetal) is a valid use case.
                # return and let native_static_libs list empty
                return
            if self.info.system == 'emscripten':
                # no match and emscripten is valid after rustc 1.84
                return
            raise EnvironmentException('Failed to find native-static-libs in Rust compiler output.')
        # Exclude some well known libraries that we don't need because they
        # are always part of C/C++ linkers. Rustc probably should not print
        # them, pkg-config for example never specify them.
        # FIXME: https://github.com/rust-lang/rust/issues/55120
        exclude = {'-lc', '-lgcc_s', '-lkernel32', '-ladvapi32', '/defaultlib:msvcrt'}
        self.native_static_libs = [i for i in match.group(1).split() if i not in exclude]

    def get_dependency_gen_args(self, outtarget: str, outfile: str) -> T.List[str]:
        return ['--emit', f'dep-info={outfile}']

    def get_output_args(self, outputname: str) -> T.List[str]:
        return ['--emit', f'link={outputname}']

    @functools.lru_cache(maxsize=None)
    def get_sysroot(self) -> str:
        cmd = self.get_exelist(ccache=False) + ['--print', 'sysroot']
        p, stdo, stde = Popen_safe_logged(cmd)
        return stdo.split('\n', maxsplit=1)[0]

    @functools.lru_cache(maxsize=None)
    def get_target_libdir(self) -> str:
        cmd = self.get_exelist(ccache=False) + ['--print', 'target-libdir']
        p, stdo, stde = Popen_safe_logged(cmd)
        return stdo.split('\n', maxsplit=1)[0]

    @functools.lru_cache(maxsize=None)
    def get_cfgs(self) -> T.List[str]:
        cmd = self.get_exelist(ccache=False) + ['--print', 'cfg']
        p, stdo, stde = Popen_safe_logged(cmd)
        return stdo.splitlines()

    @functools.lru_cache(maxsize=None)
    def get_crt_static(self) -> bool:
        return 'target_feature="crt-static"' in self.get_cfgs()

    @functools.lru_cache(maxsize=None)
    def has_verbatim(self) -> bool:
        if version_compare(self.version, '< 1.67.0'):
            return False
        # GNU ld support '-l:PATH'
        if 'ld.' in self.linker.id and self.linker.id != 'ld.wasm':
            return True
        # -l:+verbatim does not work (yet?) with MSVC link or Apple ld64
        # (https://github.com/rust-lang/rust/pull/138753).  For ld64, it
        # works together with -l:+whole_archive because -force_load (the macOS
        # equivalent of --whole-archive), receives the full path to the library
        # being linked.  However, Meson uses "bundle", not "whole_archive".
        return False

    def lib_file_to_l_arg(self, env: Environment, libname: str) -> T.Optional[str]:
        """Undo the effects of -l on the filename, returning the
           argument that can be passed to -l, or None if the
           library name is not supported."""
        if not is_library(libname):
            return None
        libname, ext = os.path.splitext(libname)

        # On Windows, rustc's -lfoo searches either foo.lib or libfoo.a.
        # Elsewhere, it searches both static and shared libraries and always with
        # the "lib" prefix; for simplicity just skip .lib on non-Windows.
        if env.machines[self.for_machine].is_windows():
            if ext == '.lib':
                return libname
            if ext != '.a':
                return None
        else:
            if ext == '.lib':
                return None

        if not libname.startswith('lib'):
            return None
        libname = libname[3:]
        return libname

    def get_debug_args(self, is_debug: bool) -> T.List[str]:
        return clike_debug_args[is_debug]

    def get_optimization_args(self, optimization_level: str) -> T.List[str]:
        return rust_optimization_args[optimization_level]

    def build_rpath_args(self, env: Environment, build_dir: str, from_dir: str,
                         target: BuildTarget, extra_paths: T.Optional[T.List[str]] = None) -> T.Tuple[T.List[str], T.Set[bytes]]:
        # add rustc's sysroot to account for rustup installations
        args, to_remove = super().build_rpath_args(env, build_dir, from_dir, target, [self.get_target_libdir()])

        rustc_rpath_args = []
        for arg in args:
            rustc_rpath_args.append('-C')
            rustc_rpath_args.append(f'link-arg={arg}')
        return rustc_rpath_args, to_remove

    def compute_parameters_with_absolute_paths(self, parameter_list: T.List[str],
                                               build_dir: str) -> T.List[str]:
        for idx, i in enumerate(parameter_list):
            if i[:2] == '-L':
                for j in ['dependency', 'crate', 'native', 'framework', 'all']:
                    combined_len = len(j) + 3
                    if i[:combined_len] == f'-L{j}=':
                        parameter_list[idx] = i[:combined_len] + os.path.normpath(os.path.join(build_dir, i[combined_len:]))
                        break

        return parameter_list

    @classmethod
    def use_linker_args(cls, linker: str, version: str) -> T.List[str]:
        return ['-C', f'linker={linker}']

    # Rust does not have a use_linker_args because it dispatches to a gcc-like
    # C compiler for dynamic linking, as such we invoke the C compiler's
    # use_linker_args method instead.

    def get_options(self) -> MutableKeyedOptionDictType:
        opts = super().get_options()

        key = self.form_compileropt_key('std')
        opts[key] = options.UserComboOption(
            self.make_option_name(key),
            'Rust edition to use',
            'none',
            choices=['none', '2015', '2018', '2021', '2024'])

        key = self.form_compileropt_key('dynamic_std')
        opts[key] = options.UserBooleanOption(
            self.make_option_name(key),
            'Whether to link Rust build targets to a dynamic libstd',
            False)

        return opts

    def get_dependency_compile_args(self, dep: 'Dependency') -> T.List[str]:
        # Rust doesn't have dependency compile arguments so simply return
        # nothing here. Dependencies are linked and all required metadata is
        # provided by the linker flags.
        return []

    def get_option_std_args(self, target: BuildTarget, env: Environment, subproject: T.Optional[str] = None) -> T.List[str]:
        args = []
        std = self.get_compileropt_value('std', env, target, subproject)
        assert isinstance(std, str)
        if std != 'none':
            args.append('--edition=' + std)
        return args

    def get_crt_compile_args(self, crt_val: str, buildtype: str) -> T.List[str]:
        # Rust handles this for us, we don't need to do anything
        return []

    def get_crt_link_args(self, crt_val: str, buildtype: str) -> T.List[str]:
        if self.linker.id not in {'link', 'lld-link'}:
            return []
        return self.MSVCRT_ARGS[self.get_crt_val(crt_val, buildtype)]

    def get_colorout_args(self, colortype: str) -> T.List[str]:
        if colortype in {'always', 'never', 'auto'}:
            return [f'--color={colortype}']
        raise MesonException(f'Invalid color type for rust {colortype}')

    def get_linker_always_args(self) -> T.List[str]:
        args: T.List[str] = []
        # Rust is super annoying, calling -C link-arg foo does not work, it has
        # to be -C link-arg=foo
        for a in super().get_linker_always_args():
            args.extend(['-C', f'link-arg={a}'])
        return args

    def get_werror_args(self) -> T.List[str]:
        # Use -D warnings, which makes every warning not explicitly allowed an
        # error
        return ['-D', 'warnings']

    def get_warn_args(self, level: str) -> T.List[str]:
        # TODO: I'm not really sure what to put here, Rustc doesn't have warning
        return self._WARNING_LEVELS[level]

    def get_pic_args(self) -> T.List[str]:
        # relocation-model=pic is rustc's default already.
        return []

    def get_pie_args(self) -> T.List[str]:
        # Rustc currently has no way to toggle this, it's controlled by whether
        # pic is on by rustc
        return []

    def get_assert_args(self, disable: bool, env: 'Environment') -> T.List[str]:
        action = "no" if disable else "yes"
        return ['-C', f'debug-assertions={action}', '-C', 'overflow-checks=no']

    def get_rust_tool(self, name: str, env: Environment) -> T.List[str]:
        if self.rustup_run_and_args:
            rustup_exelist, args = self.rustup_run_and_args
            # do not use extend so that exelist is copied
            exelist = rustup_exelist + [name]
        else:
            exelist = [name]
            args = self.get_exe_args()

        from ..programs import find_external_program
        for prog in find_external_program(env, self.for_machine, exelist[0], exelist[0],
                                          [exelist[0]], allow_default_for_cross=False):
            exelist[0] = prog.path
            break
        else:
            return []

        return exelist + args

    def has_multi_arguments(self, args: T.List[str], env: Environment) -> T.Tuple[bool, bool]:
        return self.compiles('fn main() { std::process::exit(0) }\n', env, extra_args=args, mode=CompileCheckMode.COMPILE)

    def has_multi_link_arguments(self, args: T.List[str], env: Environment) -> T.Tuple[bool, bool]:
        args = self.linker.fatal_warnings() + args
        return self.compiles('fn main() { std::process::exit(0) }\n', env, extra_args=args, mode=CompileCheckMode.LINK)

    @functools.lru_cache(maxsize=None)
    def get_rustdoc(self, env: 'Environment') -> T.Optional[RustdocTestCompiler]:
        exelist = self.get_rust_tool('rustdoc', env)
        if not exelist:
            return None

        return RustdocTestCompiler(exelist, self.version, self.for_machine,
                                   self.is_cross, self.info, full_version=self.full_version,
                                   linker=self.linker, rustc=self)

class ClippyRustCompiler(RustCompiler):

    """Clippy is a linter that wraps Rustc.

    This just provides us a different id
    """

    id = 'clippy-driver rustc'


class RustdocTestCompiler(RustCompiler):

    """We invoke Rustdoc to run doctests.  Some of the flags
       are different from rustc and some (e.g. --emit link) are
       ignored."""

    id = 'rustdoc --test'

    def __init__(self, exelist: T.List[str], version: str, for_machine: MachineChoice,
                 is_cross: bool, info: 'MachineInfo',
                 full_version: T.Optional[str],
                 linker: T.Optional['DynamicLinker'], rustc: RustCompiler):
        super().__init__(exelist, version, for_machine,
                         is_cross, info, full_version, linker)
        self.rustc = rustc

    @functools.lru_cache(maxsize=None)
    def get_sysroot(self) -> str:
        return self.rustc.get_sysroot()

    @functools.lru_cache(maxsize=None)
    def get_target_libdir(self) -> str:
        return self.rustc.get_target_libdir()

    @functools.lru_cache(maxsize=None)
    def get_cfgs(self) -> T.List[str]:
        return self.rustc.get_cfgs()

    def get_debug_args(self, is_debug: bool) -> T.List[str]:
        return []

    def get_dependency_gen_args(self, outtarget: str, outfile: str) -> T.List[str]:
        return []

    def get_output_args(self, outputname: str) -> T.List[str]:
        return []



# SPDX-License-Identifier: Apache-2.0
# Copyright 2018 The Meson development team

from __future__ import annotations

import asyncio.subprocess
import fnmatch
import itertools
import json
import signal
import sys
from pathlib import Path

from .. import mlog
from ..compilers import lang_suffixes
from ..mesonlib import quiet_git, join_args, determine_worker_count
from ..mtest import complete_all
import typing as T

Info = T.TypeVar("Info")

async def run_with_buffered_output(cmdlist: T.List[str]) -> int:
    """Run the command in cmdlist, buffering the output so that it is
       not mixed for multiple child processes.  Kill the child on
       cancellation."""
    quoted_cmdline = join_args(cmdlist)
    p: T.Optional[asyncio.subprocess.Process] = None
    try:
        p = await asyncio.create_subprocess_exec(*cmdlist,
                                                 stdin=asyncio.subprocess.DEVNULL,
                                                 stdout=asyncio.subprocess.PIPE,
                                                 stderr=asyncio.subprocess.STDOUT)
        stdo, _ = await p.communicate()
    except FileNotFoundError as e:
        print(mlog.blue('>>>'), quoted_cmdline, file=sys.stderr)
        print(mlog.red('not found:'), e.filename, file=sys.stderr)
        return 1
    except asyncio.CancelledError:
        if p:
            p.kill()
            await p.wait()
            return p.returncode or 1
        else:
            return 0

    if stdo:
        print(mlog.blue('>>>'), quoted_cmdline, flush=True)
        sys.stdout.buffer.write(stdo)
    return p.returncode

async def _run_workers(infos: T.Iterable[Info],
                       fn: T.Callable[[Info], T.Iterable[T.Coroutine[None, None, int]]]) -> int:
    futures: T.List[asyncio.Future[int]] = []
    semaphore = asyncio.Semaphore(determine_worker_count())

    async def run_one(worker_coro: T.Coroutine[None, None, int]) -> int:
        try:
            async with semaphore:
                return await worker_coro
        except asyncio.CancelledError as e:
            worker_coro.throw(e)
            return await worker_coro

    def sigterm_handler() -> None:
        for f in futures:
            f.cancel()

    if sys.platform != 'win32':
        loop = asyncio.get_running_loop()
        loop.add_signal_handler(signal.SIGINT, sigterm_handler)
        loop.add_signal_handler(signal.SIGTERM, sigterm_handler)

    for i in infos:
        futures.extend((asyncio.ensure_future(run_one(x)) for x in fn(i)))
    if not futures:
        return 0

    try:
        await complete_all(futures)
    except BaseException:
        for f in futures:
            f.cancel()
        raise

    return max(f.result() for f in futures if f.done() and not f.cancelled())

def parse_pattern_file(fname: Path) -> T.List[str]:
    patterns = []
    try:
        with fname.open(encoding='utf-8') as f:
            for line in f:
                pattern = line.strip()
                if pattern and not pattern.startswith('#'):
                    patterns.append(pattern)
    except FileNotFoundError:
        pass
    return patterns

def all_clike_files(name: str, srcdir: Path, builddir: Path) -> T.Iterable[Path]:
    patterns = parse_pattern_file(srcdir / f'.{name}-include')
    globs: T.Union[T.List[T.List[Path]], T.List[T.Generator[Path, None, None]]]
    if patterns:
        globs = [srcdir.glob(p) for p in patterns]
    else:
        r, o = quiet_git(['ls-files'], srcdir)
        if r:
            globs = [[Path(srcdir, f) for f in o.splitlines()]]
        else:
            globs = [srcdir.glob('**/*')]
    patterns = parse_pattern_file(srcdir / f'.{name}-ignore')
    ignore = [str(builddir / '*')]
    ignore.extend([str(srcdir / p) for p in patterns])
    suffixes = set(lang_suffixes['c']).union(set(lang_suffixes['cpp']))
    suffixes.add('h')
    suffixes = {f'.{s}' for s in suffixes}
    for f in itertools.chain.from_iterable(globs):
        strf = str(f)
        if f.is_dir() or f.suffix not in suffixes or \
                any(fnmatch.fnmatch(strf, i) for i in ignore):
            continue
        yield f

def run_clang_tool(name: str, srcdir: Path, builddir: Path, fn: T.Callable[..., T.Coroutine[None, None, int]], *args: T.Any) -> int:
    if sys.platform == 'win32' and sys.version_info < (3, 8):
        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())

    def wrapper(path: Path) -> T.Iterable[T.Coroutine[None, None, int]]:
        yield fn(path, *args)
    return asyncio.run(_run_workers(all_clike_files(name, srcdir, builddir), wrapper))

def run_clang_tool_on_sources(name: str, srcdir: Path, builddir: Path, fn: T.Callable[..., T.Coroutine[None, None, int]], *args: T.Any) -> int:
    if sys.platform == 'win32' and sys.version_info < (3, 8):
        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())

    source_files = set()
    with open('meson-info/intro-targets.json', encoding='utf-8') as fp:
        targets = json.load(fp)

        for target in targets:
            for target_source in target.get('target_sources') or []:
                for source in target_source.get('sources') or []:
                    source_files.add(Path(source))

    clike_files = set(all_clike_files(name, srcdir, builddir))
    source_files = source_files.intersection(clike_files)

    def wrapper(path: Path) -> T.Iterable[T.Coroutine[None, None, int]]:
        yield fn(path, *args)
    return asyncio.run(_run_workers(source_files, wrapper))

def run_tool_on_targets(fn: T.Callable[[T.Dict[str, T.Any]],
                                       T.Iterable[T.Coroutine[None, None, int]]]) -> int:
    if sys.platform == 'win32' and sys.version_info < (3, 8):
        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())

    with open('meson-info/intro-targets.json', encoding='utf-8') as fp:
        targets = json.load(fp)
    return asyncio.run(_run_workers(targets, fn))



# versions.py -- core Portage functionality
# Copyright 1998-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = [
    "best",
    "catpkgsplit",
    "catsplit",
    "cpv_getkey",
    "cpv_getversion",
    "cpv_sort_key",
    "pkgcmp",
    "pkgsplit",
    "ververify",
    "vercmp",
]

import re
import typing
import warnings
from functools import lru_cache
from typing import Any, Optional, Union
from collections.abc import Sequence

from portage import _unicode_decode
from portage.eapi import _eapi_attrs, _get_eapi_attrs
from portage.exception import InvalidData
from portage.localization import _

_unknown_repo = "__unknown__"

# \w is [a-zA-Z0-9_]

# PMS 3.1.3: A slot name may contain any of the characters [A-Za-z0-9+_.-].
# It must not begin with a hyphen, a dot or a plus sign.
_slot = r"([\w][\w+.-]*)"

# 2.1.1 A category name may contain any of the characters [A-Za-z0-9+_.-].
# It must not begin with a hyphen, a dot or a plus sign.
_cat = r"[\w][\w+.-]*"

# 2.1.2 A package name may contain any of the characters [A-Za-z0-9+_-].
# It must not begin with a hyphen or a plus sign,
# and must not end in a hyphen followed by one or more digits.
_pkg = r"[\w][\w+-]*?"

_v = r"(\d+)((\.\d+)*)([a-z]?)((_(pre|p|beta|alpha|rc)\d*)*)"
_rev = r"\d+"
_vr = _v + "(-r(" + _rev + "))?"

_cp = "(" + _cat + "/" + _pkg + "(-" + _vr + ")?)"
_cpv = "(" + _cp + "-" + _vr + ")"
_pv = (
    "(?P<pn>"
    + _pkg
    + "(?P<pn_inval>-"
    + _vr
    + ")?)"
    + "-(?P<ver>"
    + _v
    + ")(-r(?P<rev>"
    + _rev
    + "))?"
)

ver_regexp = re.compile("^" + _vr + "$")
suffix_regexp = re.compile("^(alpha|beta|rc|pre|p)(\\d*)$")
suffix_value = {"pre": -2, "p": 0, "alpha": -4, "beta": -3, "rc": -1}
endversion_keys = ["pre", "p", "alpha", "beta", "rc"]

_slot_re_cache = {}


def _get_slot_re(eapi_attrs: _eapi_attrs) -> typing.Pattern:
    cache_key = eapi_attrs.slot_operator
    slot_re = _slot_re_cache.get(cache_key)
    if slot_re is not None:
        return slot_re

    if eapi_attrs.slot_operator:
        slot_re = _slot + r"(/" + _slot + r")?"
    else:
        slot_re = _slot

    slot_re = re.compile("^" + slot_re + "$", re.VERBOSE | re.ASCII)

    _slot_re_cache[cache_key] = slot_re
    return slot_re


_pv_re = None


def _get_pv_re(eapi_attrs: _eapi_attrs) -> typing.Pattern:
    global _pv_re
    if _pv_re is not None:
        return _pv_re

    _pv_re = re.compile(r"^" + _pv + r"$", re.VERBOSE | re.ASCII)

    return _pv_re


def ververify(myver: str, silent: int = 1) -> bool:
    if ver_regexp.match(myver):
        return True
    if not silent:
        print(_("!!! syntax error in version: %s") % myver)
    return False


@lru_cache(1024)
def vercmp(ver1: str, ver2: str, silent: int = 1) -> Optional[int]:
    """
    Compare two versions
    Example usage:
            >>> from portage.versions import vercmp
            >>> vercmp('1.0-r1','1.2-r3')
            negative number
            >>> vercmp('1.3','1.2-r3')
            positive number
            >>> vercmp('1.0_p3','1.0_p3')
            0

    @param pkg1: version to compare with (see ver_regexp in portage.versions.py)
    @type pkg1: string (example: "2.1.2-r3")
    @param pkg2: version to compare againts (see ver_regexp in portage.versions.py)
    @type pkg2: string (example: "2.1.2_rc5")
    @rtype: None or float
    @return:
    1. positive if ver1 is greater than ver2
    2. negative if ver1 is less than ver2
    3. 0 if ver1 equals ver2
    4. None if ver1 or ver2 are invalid (see ver_regexp in portage.versions.py)
    """

    if ver1 == ver2:
        return 0

    match1 = ver_regexp.match(ver1)
    match2 = ver_regexp.match(ver2)

    # checking that the versions are valid
    if not match1 or not match1.groups():
        if not silent:
            print(_("!!! syntax error in version: %s") % ver1)
        return None
    if not match2 or not match2.groups():
        if not silent:
            print(_("!!! syntax error in version: %s") % ver2)
        return None

    # building lists of the version parts before the suffix
    # first part is simple
    list1 = [int(match1.group(1))]
    list2 = [int(match2.group(1))]

    # this part would greatly benefit from a fixed-length version pattern
    if match1.group(2) or match2.group(2):
        vlist1 = match1.group(2)[1:].split(".")
        vlist2 = match2.group(2)[1:].split(".")

        for i in range(0, max(len(vlist1), len(vlist2))):
            # Implcit .0 is given a value of -1, so that 1.0.0 > 1.0, since it
            # would be ambiguous if two versions that aren't literally equal
            # are given the same value (in sorting, for example).
            if len(vlist1) <= i or len(vlist1[i]) == 0:
                list1.append(-1)
                list2.append(int(vlist2[i]))
            elif len(vlist2) <= i or len(vlist2[i]) == 0:
                list1.append(int(vlist1[i]))
                list2.append(-1)
            # Let's make life easy and use integers unless we're forced to use floats
            elif vlist1[i][0] != "0" and vlist2[i][0] != "0":
                list1.append(int(vlist1[i]))
                list2.append(int(vlist2[i]))
            # now we have to use floats so 1.02 compares correctly against 1.1
            else:
                # list1.append(float("0."+vlist1[i]))
                # list2.append(float("0."+vlist2[i]))
                # Since python floats have limited range, we multiply both
                # floating point representations by a constant so that they are
                # transformed into whole numbers. This allows the practically
                # infinite range of a python int to be exploited. The
                # multiplication is done by padding both literal strings with
                # zeros as necessary to ensure equal length.
                max_len = max(len(vlist1[i]), len(vlist2[i]))
                list1.append(int(vlist1[i].ljust(max_len, "0")))
                list2.append(int(vlist2[i].ljust(max_len, "0")))

    # and now the final letter
    # NOTE: Behavior changed in r2309 (between portage-2.0.x and portage-2.1).
    # The new behavior is 12.2.5 > 12.2b which, depending on how you look at,
    # may seem counter-intuitive. However, if you really think about it, it
    # seems like it's probably safe to assume that this is the behavior that
    # is intended by anyone who would use versions such as these.
    if len(match1.group(4)):
        list1.append(ord(match1.group(4)))
    if len(match2.group(4)):
        list2.append(ord(match2.group(4)))

    for i in range(0, max(len(list1), len(list2))):
        if len(list1) <= i:
            return -1
        if len(list2) <= i:
            return 1
        if list1[i] != list2[i]:
            a = list1[i]
            b = list2[i]
            rval = (a > b) - (a < b)
            return rval

    # main version is equal, so now compare the _suffix part
    list1 = match1.group(5).split("_")[1:]
    list2 = match2.group(5).split("_")[1:]

    for i in range(0, max(len(list1), len(list2))):
        # Implicit _p0 is given a value of -1, so that 1 < 1_p0
        if len(list1) <= i:
            s1 = ("p", "-1")
        else:
            s1 = suffix_regexp.match(list1[i]).groups()
        if len(list2) <= i:
            s2 = ("p", "-1")
        else:
            s2 = suffix_regexp.match(list2[i]).groups()
        if s1[0] != s2[0]:
            a = suffix_value[s1[0]]
            b = suffix_value[s2[0]]
            rval = (a > b) - (a < b)
            return rval
        if s1[1] != s2[1]:
            # it's possible that the s(1|2)[1] == ''
            # in such a case, fudge it.
            try:
                r1 = int(s1[1])
            except ValueError:
                r1 = 0
            try:
                r2 = int(s2[1])
            except ValueError:
                r2 = 0
            rval = (r1 > r2) - (r1 < r2)
            if rval:
                return rval

    # the suffix part is equal to, so finally check the revision
    if match1.group(9):
        r1 = int(match1.group(9))
    else:
        r1 = 0
    if match2.group(9):
        r2 = int(match2.group(9))
    else:
        r2 = 0
    rval = (r1 > r2) - (r1 < r2)
    return rval


def pkgcmp(pkg1: tuple[str, str, str], pkg2: tuple[str, str, str]) -> Optional[int]:
    """
    Compare 2 package versions created in pkgsplit format.

    Example usage:
            >>> from portage.versions import *
            >>> pkgcmp(pkgsplit('test-1.0-r1'),pkgsplit('test-1.2-r3'))
            -1
            >>> pkgcmp(pkgsplit('test-1.3'),pkgsplit('test-1.2-r3'))
            1

    @param pkg1: package to compare with
    @type pkg1: list (example: ['test', '1.0', 'r1'])
    @param pkg2: package to compare againts
    @type pkg2: list (example: ['test', '1.0', 'r1'])
    @rtype: None or integer
    @return:
            1. None if package names are not the same
            2. 1 if pkg1 is greater than pkg2
            3. -1 if pkg1 is less than pkg2
            4. 0 if pkg1 equals pkg2
    """
    if pkg1[0] != pkg2[0]:
        return None
    return vercmp("-".join(pkg1[1:]), "-".join(pkg2[1:]))


def _pkgsplit(mypkg: str, eapi: Any = None) -> Optional[tuple[str, str, str]]:
    """
    @param mypkg: pv
    @return:
    1. None if input is invalid.
    2. (pn, ver, rev) if input is pv
    """
    m = _get_pv_re(_get_eapi_attrs(eapi)).match(mypkg)
    if m is None:
        return None

    if m.group("pn_inval") is not None:
        # package name appears to have a version-like suffix
        return None

    rev = m.group("rev")
    if rev is None:
        rev = "0"
    rev = "r" + rev

    return (m.group("pn"), m.group("ver"), rev)


_cat_re = re.compile(f"^{_cat}$", re.ASCII)
_missing_cat = "null"


@lru_cache(10240)
def catpkgsplit(
    mydata: Union[str, "_pkg_str"],
    silent: int = 1,
    eapi: Any = None,
) -> Optional[tuple[str, ...]]:
    """
    Takes a Category/Package-Version-Rev and returns a list of each.

    @param mydata: Data to split
    @type mydata: string
    @param silent: suppress error messages
    @type silent: Boolean (integer)
    @rype: list
    @return:
    1.  If each exists, it returns [cat, pkgname, version, rev]
    2.  If cat is not specificed in mydata, cat will be "null"
    3.  if rev does not exist it will be '-r0'
    """
    try:
        return mydata.cpv_split
    except AttributeError:
        pass
    mysplit = mydata.split("/", 1)
    p_split = None
    if len(mysplit) == 1:
        cat = _missing_cat
        p_split = _pkgsplit(mydata, eapi=eapi)
    elif len(mysplit) == 2:
        cat = mysplit[0]
        if _cat_re.match(cat) is not None:
            p_split = _pkgsplit(mysplit[1], eapi=eapi)
    if not p_split:
        return None
    retval = (cat, p_split[0], p_split[1], p_split[2])
    return retval


class _pkg_str(str):
    """
    This class represents a cpv. It inherits from str and has attributes
    that cache results for use by functions like catpkgsplit and cpv_getkey
    which are called frequently (especially in match_from_list).  Instances
    are typically created in dbapi.cp_list() or the Atom contructor, and
    propagate from there. Generally, code that pickles these objects will
    manually convert them to a plain unicode object first.

    Instances of this class will have missing attributes for metadata that
    has not been passed into the constructor. The missing attributes are
    used to distinguish missing metadata values from undefined metadata values.
    For example, the repo attribute will be missing if the 'repository' key
    is missing from the metadata dictionary.
    """

    def __new__(
        cls,
        cpv: str,
        metadata: Optional[dict[str, Any]] = None,
        settings: Any = None,
        eapi: Any = None,
        repo: Optional[str] = None,
        slot: Optional[str] = None,
        build_time: Optional[int] = None,
        build_id: Optional[str] = None,
        file_size: Optional[int] = None,
        mtime: Optional[int] = None,
        db: Any = None,
        repoconfig: Any = None,
    ):
        return str.__new__(cls, cpv)

    def __init__(
        self,
        cpv: str,
        metadata: Optional[dict[str, Any]] = None,
        settings: Any = None,
        eapi: Any = None,
        repo: Optional[str] = None,
        slot: Optional[str] = None,
        build_time: Optional[int] = None,
        build_id: Optional[str] = None,
        file_size: Optional[int] = None,
        mtime: Optional[int] = None,
        db: Any = None,
        repoconfig: Any = None,
    ):
        from portage.repository.config import _gen_valid_repo

        if not isinstance(cpv, str):
            # Avoid TypeError from str.__init__ with PyPy.
            cpv = _unicode_decode(cpv)
        str.__init__(cpv)
        if metadata is not None:
            self.__dict__["_metadata"] = metadata
            slot = metadata.get("SLOT", slot)
            repo = metadata.get("repository", repo)
            eapi = metadata.get("EAPI", eapi)
            build_time = metadata.get("BUILD_TIME", build_time)
            file_size = metadata.get("SIZE", file_size)
            build_id = metadata.get("BUILD_ID", build_id)
            mtime = metadata.get("_mtime_", mtime)
        if settings is not None:
            self.__dict__["_settings"] = settings
        if db is not None:
            self.__dict__["_db"] = db
        if repoconfig is not None:
            self.__dict__["_repoconfig"] = repoconfig
        if eapi is not None:
            self.__dict__["eapi"] = eapi

        self.__dict__["build_time"] = self._long(build_time, 0)
        self.__dict__["file_size"] = self._long(file_size, None)
        self.__dict__["build_id"] = self._long(build_id, None)
        self.__dict__["mtime"] = self._long(mtime, None)
        self.__dict__["cpv_split"] = catpkgsplit(cpv, eapi=eapi)
        if self.cpv_split is None:
            raise InvalidData(cpv)
        self.__dict__["cp"] = self.cpv_split[0] + "/" + self.cpv_split[1]
        if self.cpv_split[-1] == "r0" and cpv[-3:] != "-r0":
            self.__dict__["version"] = "-".join(self.cpv_split[2:-1])
        else:
            self.__dict__["version"] = "-".join(self.cpv_split[2:])
        # for match_from_list introspection
        self.__dict__["cpv"] = self
        if slot is not None:
            eapi_attrs = _get_eapi_attrs(eapi)
            slot_match = _get_slot_re(eapi_attrs).match(slot)
            if slot_match is None:
                # Avoid an InvalidAtom exception when creating SLOT atoms
                self.__dict__["slot"] = "0"
                self.__dict__["sub_slot"] = "0"
                self.__dict__["slot_invalid"] = slot
            else:
                if eapi_attrs.slot_operator:
                    slot_split = slot.split("/")
                    self.__dict__["slot"] = slot_split[0]
                    if len(slot_split) > 1:
                        self.__dict__["sub_slot"] = slot_split[1]
                    else:
                        self.__dict__["sub_slot"] = slot_split[0]
                else:
                    self.__dict__["slot"] = slot
                    self.__dict__["sub_slot"] = slot

        if repo is not None:
            repo = _gen_valid_repo(repo)
            if not repo:
                repo = _unknown_repo
            self.__dict__["repo"] = repo

    def __setattr__(self, name: str, value: Any) -> None:
        raise AttributeError(
            "_pkg_str instances are immutable", self.__class__, name, value
        )

    @staticmethod
    def _long(var: Any, default: int) -> int:
        if var is not None:
            try:
                var = int(var)
            except ValueError:
                if var:
                    var = -1
                else:
                    var = default
        return var

    @property
    def stable(self) -> bool:
        try:
            return self._stable
        except AttributeError:
            try:
                settings = self._settings
            except AttributeError:
                raise AttributeError("stable")
            if not settings.local_config:
                # Since repoman uses different config instances for
                # different profiles, our local instance does not
                # refer to the correct profile.
                raise AssertionError("invalid context")
            stable = settings._isStable(self)
            self.__dict__["_stable"] = stable
            return stable


def pkgsplit(
    mypkg: str, silent: int = 1, eapi: Any = None
) -> Optional[tuple[str, str, str]]:
    """
    @param mypkg: either a pv or cpv
    @return:
    1. None if input is invalid.
    2. (pn, ver, rev) if input is pv
    3. (cp, ver, rev) if input is a cpv
    """
    catpsplit = catpkgsplit(mypkg, eapi=eapi)
    if catpsplit is None:
        return None
    cat, pn, ver, rev = catpsplit
    if cat is _missing_cat and "/" not in mypkg:
        return (pn, ver, rev)
    return (cat + "/" + pn, ver, rev)


def cpv_getkey(mycpv: Union[_pkg_str, str], eapi: Any = None) -> Optional[str]:
    """Calls catpkgsplit on a cpv and returns only the cp."""
    try:
        return mycpv.cp
    except AttributeError:
        pass
    mysplit = catpkgsplit(mycpv, eapi=eapi)
    if mysplit is not None:
        return mysplit[0] + "/" + mysplit[1]

    warnings.warn(
        "portage.versions.cpv_getkey() " + f"called with invalid cpv: '{mycpv}'",
        DeprecationWarning,
        stacklevel=2,
    )

    myslash = mycpv.split("/", 1)
    mysplit = _pkgsplit(myslash[-1], eapi=eapi)
    if mysplit is None:
        return None
    mylen = len(myslash)
    if mylen == 2:
        return myslash[0] + "/" + mysplit[0]
    return mysplit[0]


def cpv_getversion(mycpv: Union[str, _pkg_str], eapi: Any = None) -> Optional[str]:
    """Returns the v (including revision) from an cpv."""
    try:
        return mycpv.version
    except AttributeError:
        pass
    cp = cpv_getkey(mycpv, eapi=eapi)
    if cp is None:
        return None
    return mycpv[len(cp + "-") :]


def cpv_sort_key(eapi: Any = None) -> Any:
    """
    Create an object for sorting cpvs, to be used as the 'key' parameter
    in places like list.sort() or sorted(). This calls catpkgsplit() once for
    each cpv and caches the result. If a given cpv is invalid or two cpvs
    have different category/package names, then plain string (> and <)
    comparison is used.

    @rtype: key object for sorting
    @return: object for use as the 'key' parameter in places like
            list.sort() or sorted()
    """
    from portage.util import cmp_sort_key

    split_cache = {}

    def cmp_cpv(cpv1: Any, cpv2: Any) -> int:
        split1 = split_cache.get(cpv1, False)
        if split1 is False:
            split1 = None
            try:
                split1 = cpv1.cpv
            except AttributeError:
                try:
                    split1 = _pkg_str(cpv1, eapi=eapi)
                except InvalidData:
                    pass
            split_cache[cpv1] = split1

        split2 = split_cache.get(cpv2, False)
        if split2 is False:
            split2 = None
            try:
                split2 = cpv2.cpv
            except AttributeError:
                try:
                    split2 = _pkg_str(cpv2, eapi=eapi)
                except InvalidData:
                    pass
            split_cache[cpv2] = split2

        if split1 is None or split2 is None or split1.cp != split2.cp:
            return (cpv1 > cpv2) - (cpv1 < cpv2)

        return vercmp(split1.version, split2.version)

    return cmp_sort_key(cmp_cpv)


def catsplit(mydep: str) -> list[str]:
    return mydep.split("/", 1)


def best(mymatches: Sequence[Any], eapi: Any = None) -> Any:
    """Accepts None arguments; assumes matches are valid."""
    if not mymatches:
        return ""
    if len(mymatches) == 1:
        return mymatches[0]
    bestmatch = mymatches[0]
    try:
        v2 = bestmatch.version
    except AttributeError:
        v2 = _pkg_str(bestmatch, eapi=eapi).version
    for x in mymatches[1:]:
        try:
            v1 = x.version
        except AttributeError:
            v1 = _pkg_str(x, eapi=eapi).version
        if vercmp(v1, v2) > 0:
            bestmatch = x
            v2 = v1
    return bestmatch



cimport cython

cdef class TreeVisitor:
    cdef public list access_path
    cdef dict dispatch_table

    cpdef visit(self, obj)
    cdef _visit(self, obj)
    cdef find_handler(self, obj)
    cdef _visitchild(self, child, parent, attrname, idx)
    cdef dict _visitchildren(self, parent, attrs, exclude)
    cpdef visitchildren(self, parent, attrs=*, exclude=*)
    cdef _raise_compiler_error(self, child, e)

cdef class VisitorTransform(TreeVisitor):
    cdef dict _process_children(self, parent, attrs=*, exclude=*)
    cpdef visitchildren(self, parent, attrs=*, exclude=*)
    cdef list _flatten_list(self, list orig_list)
    cpdef visitchild(self, parent, str attr, idx=*)

cdef class CythonTransform(VisitorTransform):
    cdef public context
    cdef public current_directives

cdef class ScopeTrackingTransform(CythonTransform):
    cdef public scope_type
    cdef public scope_node
    cdef visit_scope(self, node, scope_type)

cdef class EnvTransform(CythonTransform):
    cdef public list env_stack

cdef class MethodDispatcherTransform(EnvTransform):
    @cython.final
    cdef _visit_binop_node(self, node)
    @cython.final
    cdef _find_handler(self, match_name, bint has_kwargs)
    @cython.final
    cdef _delegate_to_assigned_value(self, node, function, arg_list, kwargs)
    @cython.final
    cdef _dispatch_to_handler(self, node, function, arg_list, kwargs)
    @cython.final
    cdef _dispatch_to_method_handler(self, attr_name, self_arg,
                                     is_unbound_method, type_name,
                                     node, function, arg_list, kwargs)

cdef class RecursiveNodeReplacer(VisitorTransform):
    cdef public orig_node
    cdef public new_node

cdef class NodeFinder(TreeVisitor):
    cdef node
    cdef public bint found



# Copyright 2005-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import errno
import stat

import portage
from portage import os
from portage.util import writemsg
from portage.versions import _pkg_str


class BinhostHandler:
    short_desc = "Generate a metadata index for binary packages"

    @staticmethod
    def name():
        return "binhost"

    def __init__(self):
        eroot = portage.settings["EROOT"]
        self._bintree = portage.db[eroot]["bintree"]
        self._bintree.populate()
        self._pkgindex_file = self._bintree._pkgindex_file
        self._pkgindex = self._bintree._load_pkgindex()

    def _need_update(self, cpv, data):
        if "MD5" not in data:
            return True

        size = data.get("SIZE")
        if size is None:
            return True

        mtime = data.get("_mtime_")
        if mtime is None:
            return True

        pkg_path = self._bintree.getname(cpv)
        try:
            s = os.lstat(pkg_path)
        except OSError as e:
            if e.errno not in (errno.ENOENT, errno.ESTALE):
                raise
            # We can't update the index for this one because
            # it disappeared.
            return False

        try:
            if int(mtime) != s[stat.ST_MTIME]:
                return True
            if int(size) != int(s.st_size):
                return True
        except ValueError:
            return True

        return False

    def check(self, **kwargs):
        onProgress = kwargs.get("onProgress", None)
        bintree = self._bintree
        # Force reindex in case pkgdir-index-trusted is enabled.
        bintree._populate_local(reindex=True)
        bintree.populated = True
        _instance_key = bintree.dbapi._instance_key
        cpv_all = self._bintree.dbapi.cpv_all()
        cpv_all.sort()
        maxval = len(cpv_all)
        if onProgress:
            onProgress(maxval, 0)
        pkgindex = self._pkgindex
        missing = []
        stale = []
        metadata = {}
        for d in pkgindex.packages:
            cpv = _pkg_str(d["CPV"], metadata=d, settings=bintree.settings)
            d["CPV"] = cpv
            metadata[_instance_key(cpv)] = d
            if not bintree.dbapi.cpv_exists(cpv):
                stale.append(cpv)
        for i, cpv in enumerate(cpv_all):
            d = metadata.get(_instance_key(cpv))
            if not d or self._need_update(cpv, d):
                missing.append(cpv)
            if onProgress:
                onProgress(maxval, i + 1)
        errors = [f"'{cpv}' is not in Packages" for cpv in missing]
        for cpv in stale:
            errors.append(f"'{cpv}' is not in the repository")

        errors.extend(self._check_compressed_index())

        if errors:
            return (False, errors)
        return (True, None)

    def _check_compressed_index(self):
        """Check the compressed index file for consistency. Return a list of error
        messages, or an empty list if no errors were found.
        """
        errors = []
        bintree = self._bintree
        timestamps = {}
        for suffix in ("", ".gz"):
            pkgindex_file = self._pkgindex_file + suffix
            try:
                st = os.stat(pkgindex_file)
            except OSError as e:
                timestamps[suffix] = None
                if e.errno == errno.ENOENT:
                    if suffix == "":
                        errors.append(f"Missing index file: {pkgindex_file}")
                    elif (
                        suffix == ".gz"
                        and "compress-index" in bintree.settings.features
                    ):
                        errors.append(f"Missing index file: {pkgindex_file}")
                else:
                    raise
            else:
                timestamps[suffix] = st[stat.ST_MTIME]

        if "compress-index" in bintree.settings.features and len(timestamps) == 2:
            if (
                timestamps[""] is not None
                and timestamps[".gz"] is not None
                and timestamps[""] != timestamps[".gz"]
            ):
                errors.append(
                    f"Uncompressed index timestamp '{timestamps['']}' is not equal to compressed index timestamp '{timestamps['.gz']}'"
                )

        if "compress-index" not in bintree.settings.features:
            if timestamps[".gz"] is not None:
                errors.append(
                    f"Compressed index exists but 'compress-index' feature is disabled: {self._pkgindex_file}.gz"
                )

        return errors

    def fix(self, **kwargs):
        onProgress = kwargs.get("onProgress", None)
        bintree = self._bintree
        # Force reindex in case pkgdir-index-trusted is enabled.
        bintree._populate_local(reindex=True)
        bintree.populated = True
        _instance_key = bintree.dbapi._instance_key
        cpv_all = self._bintree.dbapi.cpv_all()
        cpv_all.sort()
        maxval = 0
        if onProgress:
            onProgress(maxval, 0)
        pkgindex = self._pkgindex
        missing = []
        stale = []
        metadata = {}
        for d in pkgindex.packages:
            cpv = _pkg_str(d["CPV"], metadata=d, settings=bintree.settings)
            d["CPV"] = cpv
            metadata[_instance_key(cpv)] = d
            if not bintree.dbapi.cpv_exists(cpv):
                stale.append(cpv)

        for cpv in cpv_all:
            d = metadata.get(_instance_key(cpv))
            if not d or self._need_update(cpv, d):
                missing.append(cpv)

        if missing or stale or self._check_compressed_index():
            from portage import locks

            pkgindex_lock = locks.lockfile(self._pkgindex_file, wantnewlockfile=1)
            try:
                # Repopulate with lock held. If _populate_local returns
                # data then use that, since _load_pkgindex would return
                # stale data in this case.
                self._pkgindex = pkgindex = (
                    bintree._populate_local() or bintree._load_pkgindex()
                )
                cpv_all = self._bintree.dbapi.cpv_all()
                cpv_all.sort()

                # Recount stale/missing packages, with lock held.
                missing = []
                stale = []
                metadata = {}
                for d in pkgindex.packages:
                    cpv = _pkg_str(d["CPV"], metadata=d, settings=bintree.settings)
                    d["CPV"] = cpv
                    metadata[_instance_key(cpv)] = d
                    if not bintree.dbapi.cpv_exists(cpv):
                        stale.append(cpv)

                for cpv in cpv_all:
                    d = metadata.get(_instance_key(cpv))
                    if not d or self._need_update(cpv, d):
                        missing.append(cpv)

                maxval = len(missing)
                for i, cpv in enumerate(missing):
                    d = bintree._pkgindex_entry(cpv)
                    try:
                        bintree._eval_use_flags(cpv, d)
                    except portage.exception.InvalidDependString:
                        writemsg(
                            f"!!! Invalid binary package: '{bintree.getname(cpv)}'\n",
                            noiselevel=-1,
                        )
                    else:
                        metadata[_instance_key(cpv)] = d

                    if onProgress:
                        onProgress(maxval, i + 1)

                for cpv in stale:
                    del metadata[_instance_key(cpv)]

                # We've updated the pkgindex, so set it to
                # repopulate when necessary.
                bintree.populated = False

                del pkgindex.packages[:]
                pkgindex.packages.extend(metadata.values())
                bintree._update_pkgindex_header(self._pkgindex.header)
                bintree._pkgindex_write(self._pkgindex)

            finally:
                locks.unlockfile(pkgindex_lock)

        if onProgress:
            if maxval == 0:
                maxval = 1
            onProgress(maxval, maxval)
        return (True, None)



# Copyright 2019-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import functools
import io
import multiprocessing
import shlex
import signal
import tempfile
import types

import portage
from portage import shutil, os
from portage.checksum import checksum_str
from portage.const import BASH_BINARY, MANIFEST2_HASH_DEFAULTS, PORTAGE_PYM_PATH
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.tests.util.test_socks5 import AsyncHTTPServer
from portage.util.configparser import ConfigParserError
from portage.util.futures import asyncio
from portage.util.futures.executor.fork import ForkExecutor
from portage.util._async.SchedulerInterface import SchedulerInterface
from portage.util._eventloop.global_event_loop import global_event_loop
from portage.package.ebuild.config import config
from portage.package.ebuild.digestgen import digestgen
from portage.package.ebuild.fetch import (
    ContentHashLayout,
    DistfileName,
    _download_suffix,
    fetch,
    FilenameHashLayout,
    FlatLayout,
    MirrorLayoutConfig,
)
from portage._emirrordist.Config import Config as EmirrordistConfig
from _emerge.EbuildFetcher import EbuildFetcher
from _emerge.Package import Package


class EbuildFetchTestCase(TestCase):

    async def _test_interrupt(self, loop, server, async_fetch, pkg, ebuild_path):
        """Test interrupt, with server responses temporarily paused."""
        server.pause()
        pr, pw = multiprocessing.Pipe(duplex=False)
        timeout = loop.create_future()
        loop.add_reader(pr.fileno(), lambda: timeout.done() or timeout.set_result(None))
        self.assertEqual(
            await async_fetch(
                pkg,
                ebuild_path,
                timeout=timeout,
                pre_exec=functools.partial(self._pre_exec_interrupt_patch, pw),
            ),
            -signal.SIGTERM,
        )
        loop.remove_reader(pr.fileno())
        pw.close()

        # Read pid written by _async_spawn_fetch_pre_wait hook (the
        # corresponding write served to trigger the timeout above).
        pid = pr.recv()

        # Read pid written by _async_spawn_fetch_post_terminate hook,
        # in order to know when the ProcessLookupError test should
        # succeed.
        pid = pr.recv()
        pr.close()

        # Poll the process table until the pid has disappeared,
        # and fail if a short timeout expires.
        tries = 10
        while tries:
            tries -= 1

            msg = None
            if tries <= 0:
                try:
                    with open(f"/proc/{pid}/status") as f:
                        for line in f:
                            if line.startswith("State:"):
                                msg = line
                                break
                except OSError:
                    pass

            try:
                with self.assertRaises(ProcessLookupError, msg=msg):
                    os.kill(pid, 0)
            except Exception:
                if tries <= 0:
                    raise
                await asyncio.sleep(0.1)
            else:
                break

        server.resume()

    @staticmethod
    def _pre_exec_interrupt_patch(pw):
        portage.package.ebuild.fetch._async_spawn_fetch_pre_wait = functools.partial(
            EbuildFetchTestCase._fetch_pre_wait,
            pw,
        )
        portage.package.ebuild.fetch._async_spawn_fetch_post_terminate = (
            functools.partial(
                EbuildFetchTestCase._fetch_post_terminate,
                pw,
            )
        )

    @staticmethod
    def _fetch_pre_wait(pw, proc):
        pw.send(proc.pid)

    @staticmethod
    def _fetch_post_terminate(pw, proc):
        pw.send(proc.pid)

    def testEbuildFetch(self):
        user_config = {
            "make.conf": ('GENTOO_MIRRORS="{scheme}://{host}:{port}"',),
        }

        distfiles = {
            "bar": b"bar\n",
            "foo": b"foo\n",
        }

        ebuilds = {
            "dev-libs/A-1": {
                "EAPI": "7",
                "SRC_URI": """{scheme}://{host}:{port}/distfiles/bar.txt -> bar
					{scheme}://{host}:{port}/distfiles/foo.txt -> foo""",
            },
        }

        loop = SchedulerInterface(global_event_loop())

        scheme = "http"
        host = "127.0.0.1"
        content = {}

        with AsyncHTTPServer(host, content, loop) as server:
            ebuilds_subst = {}
            for cpv, metadata in ebuilds.items():
                metadata = metadata.copy()
                metadata["SRC_URI"] = metadata["SRC_URI"].format(
                    scheme=scheme, host=host, port=server.server_port
                )
                ebuilds_subst[cpv] = metadata

            user_config_subst = user_config.copy()
            for configname, configdata in user_config.items():
                configdata_sub = []
                for line in configdata:
                    configdata_sub.append(
                        line.format(scheme=scheme, host=host, port=server.server_port)
                    )
                user_config_subst[configname] = tuple(configdata_sub)

            playground = ResolverPlayground(
                ebuilds=ebuilds_subst,
                distfiles=distfiles,
                user_config=user_config_subst,
            )
            ro_distdir = tempfile.mkdtemp()
            try:
                self._testEbuildFetch(
                    loop,
                    scheme,
                    host,
                    distfiles,
                    ebuilds,
                    content,
                    server,
                    playground,
                    ro_distdir,
                )
            finally:
                shutil.rmtree(ro_distdir)
                playground.cleanup()

    def _testEbuildFetch(
        self,
        loop,
        scheme,
        host,
        orig_distfiles,
        ebuilds,
        content,
        server,
        playground,
        ro_distdir,
    ):
        mirror_layouts = (
            (
                "[structure]",
                "0=filename-hash BLAKE2B 8",
                "1=flat",
            ),
            (
                "[structure]",
                "1=filename-hash BLAKE2B 8",
                "0=flat",
            ),
            (
                "[structure]",
                "0=content-hash SHA512 8:8:8",
                "1=flat",
            ),
        )

        fetchcommand = shlex.split(playground.settings["FETCHCOMMAND"])
        fetch_bin = portage.process.find_binary(fetchcommand[0])
        if fetch_bin is None:
            self.skipTest(
                f"FETCHCOMMAND not found: {playground.settings['FETCHCOMMAND']}"
            )
        eubin = os.path.join(playground.eprefix, "usr", "bin")
        os.symlink(fetch_bin, os.path.join(eubin, os.path.basename(fetch_bin)))
        resumecommand = shlex.split(playground.settings["RESUMECOMMAND"])
        resume_bin = portage.process.find_binary(resumecommand[0])
        if resume_bin is None:
            self.skipTest(
                f"RESUMECOMMAND not found: {playground.settings['RESUMECOMMAND']}"
            )
        if resume_bin != fetch_bin:
            os.symlink(resume_bin, os.path.join(eubin, os.path.basename(resume_bin)))
        root_config = playground.trees[playground.eroot]["root_config"]
        portdb = root_config.trees["porttree"].dbapi

        def run_async(func, *args, **kwargs):
            with ForkExecutor(loop=loop) as executor:
                return loop.run_until_complete(
                    loop.run_in_executor(
                        executor, functools.partial(func, *args, **kwargs)
                    )
                )

        for layout_lines in mirror_layouts:
            settings = config(clone=playground.settings)
            layout_data = "".join(f"{line}\n" for line in layout_lines)
            mirror_conf = MirrorLayoutConfig()
            mirror_conf.read_from_file(io.StringIO(layout_data))
            layouts = mirror_conf.get_all_layouts()
            content["/distfiles/layout.conf"] = layout_data.encode("utf8")
            distfiles = {}
            for k, v in orig_distfiles.items():
                filename = DistfileName(
                    k,
                    digests={
                        algo: checksum_str(v, hashname=algo)
                        for algo in MANIFEST2_HASH_DEFAULTS
                    },
                )
                distfiles[filename] = v

                # mirror path
                for layout in layouts:
                    content["/distfiles/" + layout.get_path(filename)] = v
                # upstream path
                content[f"/distfiles/{k}.txt"] = v

            shutil.rmtree(settings["DISTDIR"])
            os.makedirs(settings["DISTDIR"])
            with open(os.path.join(settings["DISTDIR"], "layout.conf"), "w") as f:
                f.write(layout_data)

            if any(isinstance(layout, ContentHashLayout) for layout in layouts):
                content_db = os.path.join(
                    playground.eprefix, "var/db/emirrordist/content.db"
                )
                os.makedirs(os.path.dirname(content_db), exist_ok=True)
                try:
                    os.unlink(content_db)
                except OSError:
                    pass
            else:
                content_db = None

            # Demonstrate that fetch preserves a stale file in DISTDIR when no digests are given.
            foo_uri = {
                "foo": (f"{scheme}://{host}:{server.server_port}/distfiles/foo",)
            }
            foo_path = os.path.join(settings["DISTDIR"], "foo")
            foo_stale_content = b"stale content\n"
            with open(foo_path, "wb") as f:
                f.write(b"stale content\n")

            self.assertTrue(
                bool(run_async(fetch, foo_uri, settings, try_mirrors=False))
            )

            with open(foo_path, "rb") as f:
                self.assertEqual(f.read(), foo_stale_content)
            with open(foo_path, "rb") as f:
                self.assertNotEqual(f.read(), distfiles["foo"])

            # Use force=True to update the stale file.
            self.assertTrue(
                bool(run_async(fetch, foo_uri, settings, try_mirrors=False, force=True))
            )

            with open(foo_path, "rb") as f:
                self.assertEqual(f.read(), distfiles["foo"])

            # Test force=True with FEATURES=skiprocheck, using read-only DISTDIR.
            # FETCHCOMMAND is set to temporarily chmod +w DISTDIR. Note that
            # FETCHCOMMAND must perform atomic rename itself due to read-only
            # DISTDIR.
            with open(foo_path, "wb") as f:
                f.write(b"stale content\n")
            orig_fetchcommand = settings["FETCHCOMMAND"]
            orig_distdir_mode = os.stat(settings["DISTDIR"]).st_mode
            temp_fetchcommand = os.path.join(eubin, "fetchcommand")
            with open(temp_fetchcommand, "w") as f:
                f.write(
                    """
					set -e
					URI=$1
					DISTDIR=$2
					FILE=$3
					trap 'chmod a-w "${DISTDIR}"' EXIT
					chmod ug+w "${DISTDIR}"
					%s
					mv -f "${DISTDIR}/${FILE}.__download__" "${DISTDIR}/${FILE}"
				"""
                    % orig_fetchcommand.replace("${FILE}", "${FILE}.__download__")
                )
            settings["FETCHCOMMAND"] = (
                '"{}" "{}" "${{URI}}" "${{DISTDIR}}" "${{FILE}}"'.format(
                    BASH_BINARY,
                    temp_fetchcommand,
                )
            )
            settings.features.add("skiprocheck")
            settings.features.remove("distlocks")
            os.chmod(settings["DISTDIR"], 0o555)
            try:
                self.assertTrue(
                    bool(
                        run_async(
                            fetch, foo_uri, settings, try_mirrors=False, force=True
                        )
                    )
                )
            finally:
                settings["FETCHCOMMAND"] = orig_fetchcommand
                os.chmod(settings["DISTDIR"], orig_distdir_mode)
                settings.features.remove("skiprocheck")
                settings.features.add("distlocks")
                os.unlink(temp_fetchcommand)

            with open(foo_path, "rb") as f:
                self.assertEqual(f.read(), distfiles["foo"])

            # Test emirrordist invocation.
            emirrordist_cmd = (
                portage._python_interpreter,
                "-b",
                "-Wd",
                os.path.join(str(self.bindir), "emirrordist"),
                "--distfiles",
                settings["DISTDIR"],
                "--config-root",
                settings["EPREFIX"],
                "--delete",
                "--repositories-configuration",
                settings.repositories.config_string(),
                "--repo",
                "test_repo",
                "--mirror",
            )

            if content_db is not None:
                emirrordist_cmd = emirrordist_cmd + (
                    "--content-db",
                    content_db,
                )

            env = settings.environ()
            env["PYTHONPATH"] = ":".join(
                filter(
                    None,
                    [PORTAGE_PYM_PATH] + os.environ.get("PYTHONPATH", "").split(":"),
                )
            )

            for k in distfiles:
                try:
                    os.unlink(os.path.join(settings["DISTDIR"], k))
                except OSError:
                    pass

            proc = loop.run_until_complete(
                asyncio.create_subprocess_exec(*emirrordist_cmd, env=env)
            )
            self.assertEqual(loop.run_until_complete(proc.wait()), 0)

            for k in distfiles:
                with open(
                    os.path.join(settings["DISTDIR"], layouts[0].get_path(k)), "rb"
                ) as f:
                    self.assertEqual(f.read(), distfiles[k])

            if content_db is not None:
                loop.run_until_complete(
                    self._test_content_db(
                        emirrordist_cmd,
                        env,
                        layouts,
                        content_db,
                        distfiles,
                        settings,
                        portdb,
                    )
                )

            config_pool = config_pool_cls(settings)

            def async_fetch(pkg, ebuild_path, pre_exec=None, timeout=None):
                fetcher = EbuildFetcher(
                    config_pool=config_pool,
                    ebuild_path=ebuild_path,
                    fetchonly=False,
                    fetchall=True,
                    pkg=pkg,
                    scheduler=loop,
                    pre_exec=pre_exec,
                )
                fetcher.start()
                waiter = fetcher.async_wait()
                if timeout is not None:
                    timeout.add_done_callback(
                        lambda timeout: waiter.done() or fetcher.cancel()
                    )
                return waiter

            for cpv in ebuilds:
                metadata = dict(
                    zip(
                        Package.metadata_keys,
                        portdb.aux_get(cpv, Package.metadata_keys),
                    )
                )

                pkg = Package(
                    built=False,
                    cpv=cpv,
                    installed=False,
                    metadata=metadata,
                    root_config=root_config,
                    type_name="ebuild",
                )

                settings.setcpv(pkg)
                ebuild_path = portdb.findname(pkg.cpv)
                portage.doebuild_environment(
                    ebuild_path, "fetch", settings=settings, db=portdb
                )

                # Test good files in DISTDIR
                for k in settings["AA"].split():
                    os.stat(os.path.join(settings["DISTDIR"], k))
                self.assertEqual(
                    loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                )
                for k in settings["AA"].split():
                    with open(os.path.join(settings["DISTDIR"], k), "rb") as f:
                        self.assertEqual(f.read(), distfiles[k])

                # Test digestgen with fetch
                os.unlink(os.path.join(os.path.dirname(ebuild_path), "Manifest"))
                for k in settings["AA"].split():
                    os.unlink(os.path.join(settings["DISTDIR"], k))
                with ForkExecutor(loop=loop) as executor:
                    self.assertTrue(
                        bool(
                            loop.run_until_complete(
                                loop.run_in_executor(
                                    executor,
                                    functools.partial(
                                        digestgen, mysettings=settings, myportdb=portdb
                                    ),
                                )
                            )
                        )
                    )
                for k in settings["AA"].split():
                    with open(os.path.join(settings["DISTDIR"], k), "rb") as f:
                        self.assertEqual(f.read(), distfiles[k])

                # Test missing files in DISTDIR
                for k in settings["AA"].split():
                    os.unlink(os.path.join(settings["DISTDIR"], k))
                self.assertEqual(
                    loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                )
                for k in settings["AA"].split():
                    with open(os.path.join(settings["DISTDIR"], k), "rb") as f:
                        self.assertEqual(f.read(), distfiles[k])

                # Test interrupt, with server responses temporarily paused.
                for k in settings["AA"].split():
                    os.unlink(os.path.join(settings["DISTDIR"], k))
                loop.run_until_complete(
                    self._test_interrupt(loop, server, async_fetch, pkg, ebuild_path)
                )

                # Test empty files in DISTDIR
                for k in settings["AA"].split():
                    file_path = os.path.join(settings["DISTDIR"], k)
                    with open(file_path, "wb") as f:
                        pass
                    self.assertEqual(os.stat(file_path).st_size, 0)
                self.assertEqual(
                    loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                )
                for k in settings["AA"].split():
                    with open(os.path.join(settings["DISTDIR"], k), "rb") as f:
                        self.assertEqual(f.read(), distfiles[k])

                # Test non-empty files containing null bytes in DISTDIR
                for k in settings["AA"].split():
                    file_path = os.path.join(settings["DISTDIR"], k)
                    with open(file_path, "wb") as f:
                        f.write(len(distfiles[k]) * b"\0")
                    self.assertEqual(os.stat(file_path).st_size, len(distfiles[k]))
                self.assertEqual(
                    loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                )
                for k in settings["AA"].split():
                    with open(os.path.join(settings["DISTDIR"], k), "rb") as f:
                        self.assertEqual(f.read(), distfiles[k])

                # Test PORTAGE_RO_DISTDIRS
                settings["PORTAGE_RO_DISTDIRS"] = f'"{ro_distdir}"'
                orig_fetchcommand = settings["FETCHCOMMAND"]
                orig_resumecommand = settings["RESUMECOMMAND"]
                try:
                    settings["FETCHCOMMAND"] = settings["RESUMECOMMAND"] = ""
                    for k in settings["AA"].split():
                        file_path = os.path.join(settings["DISTDIR"], k)
                        os.rename(file_path, os.path.join(ro_distdir, k))
                    self.assertEqual(
                        loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                    )
                    for k in settings["AA"].split():
                        file_path = os.path.join(settings["DISTDIR"], k)
                        self.assertTrue(os.path.islink(file_path))
                        with open(file_path, "rb") as f:
                            self.assertEqual(f.read(), distfiles[k])
                        os.unlink(file_path)
                finally:
                    settings.pop("PORTAGE_RO_DISTDIRS")
                    settings["FETCHCOMMAND"] = orig_fetchcommand
                    settings["RESUMECOMMAND"] = orig_resumecommand

                # Test local filesystem in GENTOO_MIRRORS
                orig_mirrors = settings["GENTOO_MIRRORS"]
                orig_fetchcommand = settings["FETCHCOMMAND"]
                try:
                    settings["GENTOO_MIRRORS"] = ro_distdir
                    settings["FETCHCOMMAND"] = settings["RESUMECOMMAND"] = ""
                    self.assertEqual(
                        loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                    )
                    for k in settings["AA"].split():
                        with open(os.path.join(settings["DISTDIR"], k), "rb") as f:
                            self.assertEqual(f.read(), distfiles[k])
                finally:
                    settings["GENTOO_MIRRORS"] = orig_mirrors
                    settings["FETCHCOMMAND"] = orig_fetchcommand
                    settings["RESUMECOMMAND"] = orig_resumecommand

                # Test readonly DISTDIR
                orig_distdir_mode = os.stat(settings["DISTDIR"]).st_mode
                try:
                    os.chmod(settings["DISTDIR"], 0o555)
                    self.assertEqual(
                        loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                    )
                    for k in settings["AA"].split():
                        with open(os.path.join(settings["DISTDIR"], k), "rb") as f:
                            self.assertEqual(f.read(), distfiles[k])
                finally:
                    os.chmod(settings["DISTDIR"], orig_distdir_mode)

                # Test parallel-fetch mode
                settings["PORTAGE_PARALLEL_FETCHONLY"] = "1"
                try:
                    self.assertEqual(
                        loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                    )
                    for k in settings["AA"].split():
                        with open(os.path.join(settings["DISTDIR"], k), "rb") as f:
                            self.assertEqual(f.read(), distfiles[k])
                    for k in settings["AA"].split():
                        os.unlink(os.path.join(settings["DISTDIR"], k))
                    self.assertEqual(
                        loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                    )
                    for k in settings["AA"].split():
                        with open(os.path.join(settings["DISTDIR"], k), "rb") as f:
                            self.assertEqual(f.read(), distfiles[k])
                finally:
                    settings.pop("PORTAGE_PARALLEL_FETCHONLY")

                # Test RESUMECOMMAND
                orig_resume_min_size = settings["PORTAGE_FETCH_RESUME_MIN_SIZE"]
                try:
                    settings["PORTAGE_FETCH_RESUME_MIN_SIZE"] = "2"
                    for k in settings["AA"].split():
                        file_path = os.path.join(settings["DISTDIR"], k)
                        os.unlink(file_path)
                        with open(file_path + _download_suffix, "wb") as f:
                            f.write(distfiles[k][:2])
                    self.assertEqual(
                        loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                    )
                    for k in settings["AA"].split():
                        with open(os.path.join(settings["DISTDIR"], k), "rb") as f:
                            self.assertEqual(f.read(), distfiles[k])
                finally:
                    settings["PORTAGE_FETCH_RESUME_MIN_SIZE"] = orig_resume_min_size

                # Test readonly DISTDIR + skiprocheck, with FETCHCOMMAND set to temporarily chmod DISTDIR
                orig_fetchcommand = settings["FETCHCOMMAND"]
                orig_distdir_mode = os.stat(settings["DISTDIR"]).st_mode
                for k in settings["AA"].split():
                    os.unlink(os.path.join(settings["DISTDIR"], k))
                try:
                    os.chmod(settings["DISTDIR"], 0o555)
                    settings["FETCHCOMMAND"] = (
                        '"%s" -c "chmod ug+w \\"${DISTDIR}\\"; %s; status=\\$?; chmod a-w \\"${DISTDIR}\\"; exit \\$status"'
                        % (BASH_BINARY, orig_fetchcommand.replace('"', '\\"'))
                    )
                    settings.features.add("skiprocheck")
                    settings.features.remove("distlocks")
                    self.assertEqual(
                        loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                    )
                finally:
                    settings["FETCHCOMMAND"] = orig_fetchcommand
                    os.chmod(settings["DISTDIR"], orig_distdir_mode)
                    settings.features.remove("skiprocheck")
                    settings.features.add("distlocks")

    async def _test_content_db(
        self, emirrordist_cmd, env, layouts, content_db, distfiles, settings, portdb
    ):
        # Simulate distfile digest change for ContentDB.
        emdisopts = types.SimpleNamespace(
            content_db=content_db, distfiles=settings["DISTDIR"]
        )
        with EmirrordistConfig(
            emdisopts, portdb, asyncio.get_event_loop()
        ) as emdisconf:
            # Copy revisions from bar to foo.
            for revision_key in emdisconf.content_db["filename:bar"]:
                emdisconf.content_db.add(
                    DistfileName("foo", digests=dict(revision_key))
                )

            # Copy revisions from foo to bar.
            for revision_key in emdisconf.content_db["filename:foo"]:
                emdisconf.content_db.add(
                    DistfileName("bar", digests=dict(revision_key))
                )

            content_db_state = dict(emdisconf.content_db.items())
            self.assertEqual(content_db_state, dict(emdisconf.content_db.items()))
            self.assertEqual(
                {
                    k[len("filename:") :]
                    for k in content_db_state
                    if k.startswith("filename:")
                },
                {"bar", "foo"},
            )
            self.assertEqual(
                content_db_state["filename:foo"], content_db_state["filename:bar"]
            )
            self.assertEqual(len(content_db_state["filename:foo"]), 2)

        for k in distfiles:
            try:
                os.unlink(os.path.join(settings["DISTDIR"], k))
            except OSError:
                pass

        proc = await asyncio.create_subprocess_exec(*emirrordist_cmd, env=env)
        self.assertEqual(await proc.wait(), 0)

        for k in distfiles:
            with open(
                os.path.join(settings["DISTDIR"], layouts[0].get_path(k)), "rb"
            ) as f:
                self.assertEqual(f.read(), distfiles[k])

        with EmirrordistConfig(
            emdisopts, portdb, asyncio.get_event_loop()
        ) as emdisconf:
            self.assertEqual(content_db_state, dict(emdisconf.content_db.items()))

            # Verify that remove works as expected
            filename = [filename for filename in distfiles if filename == "foo"][0]
            self.assertTrue(bool(filename.digests))
            emdisconf.content_db.remove(filename)
            # foo should still have a content revision corresponding to bar's content.
            self.assertEqual(
                {
                    k[len("filename:") :]
                    for k in emdisconf.content_db
                    if k.startswith("filename:")
                },
                {"bar", "foo"},
            )
            self.assertEqual(len(emdisconf.content_db["filename:foo"]), 1)
            self.assertEqual(
                len(
                    [
                        revision_key
                        for revision_key in emdisconf.content_db["filename:foo"]
                        if not filename.digests_equal(
                            DistfileName(
                                "foo",
                                digests=dict(revision_key),
                            )
                        )
                    ]
                ),
                1,
            )
            # bar should still have a content revision corresponding to foo's content.
            self.assertEqual(len(emdisconf.content_db["filename:bar"]), 2)
            self.assertEqual(
                len(
                    [
                        revision_key
                        for revision_key in emdisconf.content_db["filename:bar"]
                        if filename.digests_equal(
                            DistfileName(
                                "bar",
                                digests=dict(revision_key),
                            )
                        )
                    ]
                ),
                1,
            )
            # remove the foo which refers to bar's content
            bar = [filename for filename in distfiles if filename == "bar"][0]
            foo_remaining = DistfileName("foo", digests=bar.digests)
            emdisconf.content_db.remove(foo_remaining)
            self.assertEqual(
                [
                    k[len("filename:") :]
                    for k in emdisconf.content_db
                    if k.startswith("filename:")
                ],
                ["bar"],
            )
            self.assertRaises(
                KeyError, emdisconf.content_db.__getitem__, "filename:foo"
            )
            # bar should still have a content revision corresponding to foo's content.
            self.assertEqual(len(emdisconf.content_db["filename:bar"]), 2)

    def test_flat_layout(self):
        self.assertTrue(FlatLayout.verify_args(("flat",)))
        self.assertFalse(FlatLayout.verify_args(("flat", "extraneous-arg")))
        self.assertEqual(FlatLayout().get_path("foo-1.tar.gz"), "foo-1.tar.gz")

    def test_filename_hash_layout(self):
        self.assertFalse(FilenameHashLayout.verify_args(("filename-hash",)))
        self.assertTrue(FilenameHashLayout.verify_args(("filename-hash", "SHA1", "8")))
        self.assertFalse(
            FilenameHashLayout.verify_args(("filename-hash", "INVALID-HASH", "8"))
        )
        self.assertTrue(
            FilenameHashLayout.verify_args(("filename-hash", "SHA1", "4:8:12"))
        )
        self.assertFalse(FilenameHashLayout.verify_args(("filename-hash", "SHA1", "3")))
        self.assertFalse(
            FilenameHashLayout.verify_args(("filename-hash", "SHA1", "junk"))
        )
        self.assertFalse(
            FilenameHashLayout.verify_args(("filename-hash", "SHA1", "4:8:junk"))
        )

        self.assertEqual(
            FilenameHashLayout("SHA1", "4").get_path("foo-1.tar.gz"), "1/foo-1.tar.gz"
        )
        self.assertEqual(
            FilenameHashLayout("SHA1", "8").get_path("foo-1.tar.gz"), "19/foo-1.tar.gz"
        )
        self.assertEqual(
            FilenameHashLayout("SHA1", "8:16").get_path("foo-1.tar.gz"),
            "19/c3b6/foo-1.tar.gz",
        )
        self.assertEqual(
            FilenameHashLayout("SHA1", "8:16:24").get_path("foo-1.tar.gz"),
            "19/c3b6/37a94b/foo-1.tar.gz",
        )

    def test_content_hash_layout(self):
        self.assertFalse(ContentHashLayout.verify_args(("content-hash",)))
        self.assertTrue(ContentHashLayout.verify_args(("content-hash", "SHA1", "8")))
        self.assertFalse(
            ContentHashLayout.verify_args(("content-hash", "INVALID-HASH", "8"))
        )
        self.assertTrue(
            ContentHashLayout.verify_args(("content-hash", "SHA1", "4:8:12"))
        )
        self.assertFalse(ContentHashLayout.verify_args(("content-hash", "SHA1", "3")))
        self.assertFalse(
            ContentHashLayout.verify_args(("content-hash", "SHA1", "junk"))
        )
        self.assertFalse(
            ContentHashLayout.verify_args(("content-hash", "SHA1", "4:8:junk"))
        )

        filename = DistfileName(
            "foo-1.tar.gz",
            digests={
                algo: checksum_str(b"", hashname=algo)
                for algo in MANIFEST2_HASH_DEFAULTS
            },
        )

        # Raise KeyError for a hash algorithm SHA1 which is not in MANIFEST2_HASH_DEFAULTS.
        self.assertRaises(KeyError, ContentHashLayout("SHA1", "4").get_path, filename)

        # Raise AttributeError for a plain string argument.
        self.assertRaises(
            AttributeError, ContentHashLayout("SHA512", "4").get_path, str(filename)
        )

        self.assertEqual(
            ContentHashLayout("SHA512", "4").get_path(filename),
            "c/cf83e1357eefb8bdf1542850d66d8007d620e4050b5715dc83f4a921d36ce9ce47d0d13c5d85f2b0ff8318d2877eec2f63b931bd47417a81a538327af927da3e",
        )
        self.assertEqual(
            ContentHashLayout("SHA512", "8").get_path(filename),
            "cf/cf83e1357eefb8bdf1542850d66d8007d620e4050b5715dc83f4a921d36ce9ce47d0d13c5d85f2b0ff8318d2877eec2f63b931bd47417a81a538327af927da3e",
        )
        self.assertEqual(
            ContentHashLayout("SHA512", "8:16").get_path(filename),
            "cf/83e1/cf83e1357eefb8bdf1542850d66d8007d620e4050b5715dc83f4a921d36ce9ce47d0d13c5d85f2b0ff8318d2877eec2f63b931bd47417a81a538327af927da3e",
        )
        self.assertEqual(
            ContentHashLayout("SHA512", "8:16:24").get_path(filename),
            "cf/83e1/357eef/cf83e1357eefb8bdf1542850d66d8007d620e4050b5715dc83f4a921d36ce9ce47d0d13c5d85f2b0ff8318d2877eec2f63b931bd47417a81a538327af927da3e",
        )

    def test_mirror_layout_config(self):
        mlc = MirrorLayoutConfig()
        self.assertEqual(mlc.serialize(), ())
        self.assertIsInstance(mlc.get_best_supported_layout(), FlatLayout)

        conf = """
[structure]
0=flat
"""
        mlc.read_from_file(io.StringIO(conf))
        self.assertEqual(mlc.serialize(), (("flat",),))
        self.assertIsInstance(mlc.get_best_supported_layout(), FlatLayout)
        self.assertEqual(
            mlc.get_best_supported_layout().get_path("foo-1.tar.gz"), "foo-1.tar.gz"
        )

        conf = """
[structure]
0=filename-hash SHA1 8:16
1=flat
"""
        mlc.read_from_file(io.StringIO(conf))
        self.assertEqual(
            mlc.serialize(), (("filename-hash", "SHA1", "8:16"), ("flat",))
        )
        self.assertIsInstance(mlc.get_best_supported_layout(), FilenameHashLayout)
        self.assertEqual(
            mlc.get_best_supported_layout().get_path("foo-1.tar.gz"),
            "19/c3b6/foo-1.tar.gz",
        )
        serialized = mlc.serialize()

        # test fallback
        conf = """
[structure]
0=filename-hash INVALID-HASH 8:16
1=filename-hash SHA1 32
2=flat
"""
        mlc.read_from_file(io.StringIO(conf))
        self.assertEqual(
            mlc.serialize(),
            (
                ("filename-hash", "INVALID-HASH", "8:16"),
                ("filename-hash", "SHA1", "32"),
                ("flat",),
            ),
        )
        self.assertIsInstance(mlc.get_best_supported_layout(), FilenameHashLayout)
        self.assertEqual(
            mlc.get_best_supported_layout().get_path("foo-1.tar.gz"),
            "19c3b637/foo-1.tar.gz",
        )

        # test deserialization
        mlc.deserialize(serialized)
        self.assertEqual(
            mlc.serialize(), (("filename-hash", "SHA1", "8:16"), ("flat",))
        )
        self.assertIsInstance(mlc.get_best_supported_layout(), FilenameHashLayout)
        self.assertEqual(
            mlc.get_best_supported_layout().get_path("foo-1.tar.gz"),
            "19/c3b6/foo-1.tar.gz",
        )

        # test erraneous input
        conf = """
[#(*DA*&*F
[structure]
0=filename-hash SHA1 32
"""
        self.assertRaises(ConfigParserError, mlc.read_from_file, io.StringIO(conf))

    def test_filename_hash_layout_get_filenames(self):
        filename = DistfileName(
            "foo-1.tar.gz",
            digests={
                algo: checksum_str(b"", hashname=algo)
                for algo in MANIFEST2_HASH_DEFAULTS
            },
        )
        layouts = (
            FlatLayout(),
            FilenameHashLayout("SHA1", "4"),
            FilenameHashLayout("SHA1", "8"),
            FilenameHashLayout("SHA1", "8:16"),
            FilenameHashLayout("SHA1", "8:16:24"),
            ContentHashLayout("SHA512", "8:8:8"),
        )

        for layout in layouts:
            distdir = tempfile.mkdtemp()
            try:
                path = os.path.join(distdir, layout.get_path(filename))
                try:
                    os.makedirs(os.path.dirname(path))
                except OSError:
                    pass

                with open(path, "wb") as f:
                    pass

                file_list = list(layout.get_filenames(distdir))
                self.assertTrue(len(file_list) > 0)
                for filename_result in file_list:
                    if isinstance(filename_result, DistfileName):
                        self.assertTrue(filename_result.digests_equal(filename))
                    else:
                        self.assertEqual(filename_result, str(filename))
            finally:
                shutil.rmtree(distdir)


# Tests only work with one ebuild at a time, so the config
# pool only needs a single config instance.
class config_pool_cls:
    def __init__(self, settings):
        self._settings = settings

    def allocate(self):
        return self._settings

    def deallocate(self, settings):
        pass



# Copyright 1999-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import io
import sys

from _emerge.BinpkgFetcher import BinpkgFetcher
from _emerge.CompositeTask import CompositeTask
from _emerge.BinpkgVerifier import BinpkgVerifier
import portage
from portage import os
from portage.elog import messages as elog_messages
from portage.util import no_color


class BinpkgPrefetcher(CompositeTask):
    __slots__ = ("pkg",) + (
        "pkg_path",
        "pkg_allocated_path",
        "_bintree",
    )

    def _start(self):
        self._bintree = self.pkg.root_config.trees["bintree"]
        fetcher = BinpkgFetcher(
            background=self.background,
            logfile=self.scheduler.fetch.log_file,
            pkg=self.pkg,
            scheduler=self.scheduler,
        )
        self.pkg_path = fetcher.pkg_path
        self.pkg_allocated_path = fetcher.pkg_allocated_path
        self._start_task(fetcher, self._fetcher_exit)

    def _fetcher_exit(self, fetcher):
        if self._default_exit(fetcher) != os.EX_OK:
            self.wait()
            return

        verifier = BinpkgVerifier(
            background=self.background,
            logfile=self.scheduler.fetch.log_file,
            pkg=self.pkg,
            scheduler=self.scheduler,
            _pkg_path=self.pkg_path,
        )
        self._start_task(verifier, self._verifier_exit)

    def _verifier_exit(self, verifier):
        if self._default_exit(verifier) != os.EX_OK:
            self.wait()
            return

        if self._bintree.get_local_repo_location(self.pkg.cpv):
            os.rename(self.pkg_path, self.pkg_allocated_path)
            self._current_task = None
            self.returncode = os.EX_OK
            self.wait()
            return

        injected_pkg = None
        stdout_orig = sys.stdout
        stderr_orig = sys.stderr
        out = io.StringIO()
        try:
            sys.stdout = out
            sys.stderr = out

            injected_pkg = self._bintree.inject(
                self.pkg.cpv,
                current_pkg_path=self.pkg_path,
                allocated_pkg_path=self.pkg_allocated_path,
            )

        finally:
            sys.stdout = stdout_orig
            sys.stderr = stderr_orig

            output_value = out.getvalue()
            if output_value:
                if injected_pkg is None:
                    msg = ["Binary package is not usable:"]
                    msg.extend("\t" + line for line in output_value.splitlines())
                    self._elog("eerror", msg)
                else:
                    self.scheduler.output(
                        output_value,
                        log_path=self.scheduler.fetch.log_file,
                        background=self.background,
                    )

        self._current_task = None
        self.returncode = 1 if injected_pkg is None else os.EX_OK
        self.wait()

    def _elog(self, elog_funcname, lines, phase="other"):
        out = io.StringIO()
        elog_func = getattr(elog_messages, elog_funcname)
        global_havecolor = portage.output.havecolor
        try:
            portage.output.havecolor = not no_color(self._bintree.settings)
            for line in lines:
                elog_func(line, phase=phase, key=self.pkg.cpv, out=out)
        finally:
            portage.output.havecolor = global_havecolor
        msg = out.getvalue()
        if msg:
            self.scheduler.output(
                msg, background=self.background, log_path=self.scheduler.fetch.log_file
            )



# Copyright 2020-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage import os
from portage.tests import TestCase
from portage.util._async.PipeLogger import PipeLogger
from portage.util.futures import asyncio
from portage.util.futures._asyncio.streams import _reader, _writer
from portage.util.futures.unix_events import _set_nonblocking


class PipeLoggerTestCase(TestCase):
    async def _testPipeLoggerToPipe(self, test_string, loop):
        """
        Test PipeLogger writing to a pipe connected to a PipeReader.
        This verifies that PipeLogger does not deadlock when writing
        to a pipe that's drained by a PipeReader running in the same
        process (requires non-blocking write).
        """

        input_fd, writer_pipe = os.pipe()
        _set_nonblocking(writer_pipe)
        writer_pipe = os.fdopen(writer_pipe, "wb", 0)
        writer = asyncio.ensure_future(
            _writer(writer_pipe, test_string.encode("ascii"))
        )
        writer.add_done_callback(lambda writer: writer_pipe.close())

        pr, pw = os.pipe()

        consumer = PipeLogger(
            background=True,
            input_fd=input_fd,
            log_file_path=os.fdopen(pw, "wb", 0),
            scheduler=loop,
        )
        consumer.start()

        # Before starting the reader, wait here for a moment, in order
        # to exercise PipeLogger's handling of EAGAIN during write.
        await asyncio.wait([writer], timeout=0.01)

        reader = _reader(pr)
        await writer
        content = await reader
        await consumer.async_wait()

        self.assertEqual(consumer.returncode, os.EX_OK)

        return content.decode("ascii", "replace")

    def testPipeLogger(self):
        loop = asyncio._wrap_loop()

        for x in (
            1,
            2,
            5,
            6,
            7,
            8,
            2**5,
            2**10,
            2**12,
            2**13,
            2**14,
            2**17,
            2**17 + 1,
        ):
            test_string = x * "a"
            output = loop.run_until_complete(
                self._testPipeLoggerToPipe(test_string, loop)
            )
            self.assertEqual(
                test_string, output, f"x = {x}, len(output) = {len(output)}"
            )



# Copyright 2011-2021, 2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

"""This module defines a baseline for portage's functionality.

Multiple portage commands are executed in a sequence in a playground
(see the ``baseline_command`` fixture in ``conftest.py``).

All the commands are triggered from the ``test_portage_baseline`` test.
That test is marked with::

  @pytest.mark.ft

so that it can selected with that marker, i.e.::

  pytest -m ft

``ft`` stands for *functional test*, since that's what it is, a
functional or end-to-end test, ensuring some functionality of portage.

The test also works with pytest-xdist, e.g.::

  pytest -m ft -n 8

"""

import subprocess

import pytest

import portage
from portage import os
from portage.const import (
    PORTAGE_PYM_PATH,
    USER_CONFIG_PATH,
)
from portage.process import find_binary
from portage.tests import cnf_etc_path
from portage.util import ensure_dirs
from portage.util.futures import asyncio


_METADATA_XML_FILES = (
    (
        "dev-libs/A",
        {
            "flags": "<flag name='flag'>Description of how USE='flag' affects this package</flag>",
        },
    ),
    (
        "dev-libs/B",
        {
            "flags": "<flag name='flag'>Description of how USE='flag' affects this package</flag>",
        },
    ),
)

_1Q_2010_UPDATE = """
slotmove =app-doc/pms-3 2 3
move dev-util/git dev-vcs/git
"""


@pytest.mark.ft
def test_portage_baseline(async_loop, playground, binhost, baseline_command):
    async_loop.run_until_complete(
        asyncio.ensure_future(
            _async_test_baseline(
                playground,
                binhost,
                baseline_command,
            ),
            loop=async_loop,
        )
    )


async def _async_test_baseline(playground, binhost, commands):
    debug = playground.debug
    settings = playground.settings
    trees = playground.trees
    eprefix = settings["EPREFIX"]

    test_repo_location = settings.repositories["test_repo"].location
    var_cache_edb = os.path.join(eprefix, "var", "cache", "edb")
    cachedir = os.path.join(var_cache_edb, "dep")
    cachedir_pregen = os.path.join(test_repo_location, "metadata", "md5-cache")

    cross_prefix = os.path.join(eprefix, "cross_prefix")
    cross_root = os.path.join(eprefix, "cross_root")
    cross_eroot = os.path.join(cross_root, eprefix.lstrip(os.sep))

    distdir = playground.distdir
    pkgdir = playground.pkgdir
    fake_bin = os.path.join(eprefix, "bin")
    portage_tmpdir = os.path.join(eprefix, "var", "tmp", "portage")
    profile_path = settings.profile_path
    user_config_dir = os.path.join(os.sep, eprefix, USER_CONFIG_PATH)

    path = settings.get("PATH")
    if path is not None and not path.strip():
        path = None
    if path is None:
        path = ""
    else:
        path = ":" + path
    path = fake_bin + path

    pythonpath = os.environ.get("PYTHONPATH")
    if pythonpath is not None and not pythonpath.strip():
        pythonpath = None
    if pythonpath is not None and pythonpath.split(":")[0] == PORTAGE_PYM_PATH:
        pass
    else:
        if pythonpath is None:
            pythonpath = ""
        else:
            pythonpath = ":" + pythonpath
        pythonpath = PORTAGE_PYM_PATH + pythonpath

    env = {
        "PORTAGE_OVERRIDE_EPREFIX": eprefix,
        "CLEAN_DELAY": "0",
        "DISTDIR": distdir,
        "EMERGE_WARNING_DELAY": "0",
        "INFODIR": "",
        "INFOPATH": "",
        "PATH": path,
        "PKGDIR": pkgdir,
        "PORTAGE_INST_GID": str(os.getgid()),  # str(portage.data.portage_gid),
        "PORTAGE_INST_UID": str(os.getuid()),  # str(portage.data.portage_uid),
        "PORTAGE_PYTHON": portage._python_interpreter,
        "PORTAGE_REPOSITORIES": settings.repositories.config_string(),
        "PORTAGE_TMPDIR": portage_tmpdir,
        "PORTAGE_LOGDIR": portage_tmpdir,
        "PYTHONDONTWRITEBYTECODE": os.environ.get("PYTHONDONTWRITEBYTECODE", ""),
        "PYTHONPATH": pythonpath,
        "__PORTAGE_TEST_PATH_OVERRIDE": fake_bin,
    }

    if "__PORTAGE_TEST_HARDLINK_LOCKS" in os.environ:
        env["__PORTAGE_TEST_HARDLINK_LOCKS"] = os.environ[
            "__PORTAGE_TEST_HARDLINK_LOCKS"
        ]

    updates_dir = os.path.join(test_repo_location, "profiles", "updates")
    dirs = [
        cachedir,
        cachedir_pregen,
        cross_eroot,
        cross_prefix,
        distdir,
        fake_bin,
        portage_tmpdir,
        updates_dir,
        user_config_dir,
        var_cache_edb,
    ]
    etc_symlinks = ("dispatch-conf.conf", "etc-update.conf")
    # Override things that may be unavailable, or may have portability
    # issues when running tests in exotic environments.
    #   prepstrip - bug #447810 (bash read builtin EINTR problem)
    true_symlinks = ["find", "prepstrip", "sed", "scanelf"]
    true_binary = find_binary("true")
    assert true_binary is not None, "true command not found"

    for d in dirs:
        ensure_dirs(d)
    for x in true_symlinks:
        try:
            os.symlink(true_binary, os.path.join(fake_bin, x))
        except FileExistsError:
            pass
    for x in etc_symlinks:
        try:
            os.symlink(
                os.path.join(str(cnf_etc_path), x), os.path.join(eprefix, "etc", x)
            )
        except FileExistsError:
            pass
    with open(os.path.join(var_cache_edb, "counter"), "wb") as f:
        f.write(b"100")
    # non-empty system set keeps --depclean quiet
    with open(os.path.join(profile_path, "packages"), "w") as f:
        f.write("*dev-libs/token-system-pkg")
    for cp, xml_data in _METADATA_XML_FILES:
        with open(os.path.join(test_repo_location, cp, "metadata.xml"), "w") as f:
            f.write(playground.metadata_xml_template % xml_data)
        with open(os.path.join(updates_dir, "1Q-2010"), "w") as f:
            f.write(_1Q_2010_UPDATE)
    if debug:
        # The subprocess inherits both stdout and stderr, for
        # debugging purposes.
        stdout = None
    else:
        # The subprocess inherits stderr so that any warnings
        # triggered by python -Wd will be visible.
        stdout = subprocess.PIPE

    for command in commands:
        if command:
            command.base_environment = env

            proc = await asyncio.create_subprocess_exec(
                *command(), env=command.env, stderr=None, stdout=stdout
            )

            if debug:
                await proc.wait()
            else:
                output, _err = await proc.communicate()
                await proc.wait()
                if proc.returncode != os.EX_OK:
                    portage.writemsg(output)

            real_command = command.name
            args = command.args
            assert (
                os.EX_OK == proc.returncode
            ), f"'{real_command}' failed with args '{args}'"
            command.check_command_result()



# Copyright 2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.dep import Atom
from portage.const import LIBC_PACKAGE_ATOM
from portage.dbapi._expand_new_virt import expand_new_virt

import portage.dbapi.porttree


def find_libc_deps(portdb: portage.dbapi.porttree.dbapi, realized: bool = False):
    """Finds libc package for a ROOT via portdb.

    Parameters
    ----------
    portdb : dbapi
        dbapi instance for portdb (for installed packages).
    realized : bool
        Request installed atoms rather than the installed package satisfying LIBC_PACKAGE_ATOM.

    Returns
    -------
    list
        List of libc packages (or atoms if realized is passed).
    """

    libc_pkgs = set()

    for atom in expand_new_virt(
        portdb,
        LIBC_PACKAGE_ATOM,
    ):
        if atom.blocker:
            continue

        if not realized:
            # Just the raw packages were requested (whatever satifies the virtual)
            libc_pkgs.add(atom)
            continue

        # This will give us something like sys-libs/glibc:2.2, but we want to know
        # what installed atom actually satifies that.
        try:
            libc_pkgs.add(portdb.match(atom)[0])
        except IndexError:
            continue

    return libc_pkgs


def strip_libc_deps(dep_struct: list, libc_deps: set):
    """Strip libc dependency out of a given dependency strucutre.

    Parameters
    ----------
    dep_struct: list
        List of package dependencies (atoms).

    libc_deps: set
        List of dependencies satisfying LIBC_PACKAGE_ATOM to be
        stripped out of any dependencies.

    Returns
    -------
    list
        List of dependencies with any matching libc_deps removed.
    """
    # We're going to just grab the libc provider for ROOT and
    # strip out any dep for the purposes of --changed-deps.
    # We can't go off versions, even though it'd be more precise
    # (see below), because we'd end up with FPs and unnecessary
    # --changed-deps results far too often.
    #
    # This penalizes a bit the case where someone adds a
    # minimum (or maximum) version of libc explicitly in an ebuild
    # without a new revision, but that's extremely rare, and doesn't
    # feel like it changes the balance for what we prefer here.

    for i, x in reversed(list(enumerate(dep_struct))):
        # We only need to bother if x is an Atom because we know the deps
        # we inject are simple & flat.
        if isinstance(x, Atom) and any(x.cp == libc_dep.cp for libc_dep in libc_deps):
            del dep_struct[i]
